from enum import Enum
from marshmallow import EXCLUDE, fields, validate

from aries_cloudagent.messaging.agent_message import AgentMessage, AgentMessageSchema
from aries_cloudagent.messaging.valid import UUIDFour

from ..message_types import (
    PROTOCOL_PACKAGE,
    DATA_AGREEMENT_QR_CODE_WORKFLOW_PROBLEM_REPORT
)

# Handler class path for Data Agreement Qr code workflow
# Problem Report (data-agreement-qr-code/1.0/problem-report) message
DATA_AGREEMENT_QR_CODE_WORKFLOW_PROBLEM_REPORT_HANDLER_CLASS = (
    f"{PROTOCOL_PACKAGE}.handlers"
    ".data_agreement_qr_code_problem_report_handler.DataAgreementQrCodeProblemReportHandler"
)


class DataAgreementQrCodeProblemReportReason(str, Enum):
    """Supported reason codes."""

    # Trigger when qr code identifier provided is invalid.
    INVALID_QR_ID = "invalid_qr_id"

    # Trigger QR code is already scanned.
    QR_CODE_SCANNED_ONCE = "qr_code_scanned_once"

    # Failed to process qr code workflow initiate message due to an internal error.
    FAILED_TO_PROCESS_QR_CODE_INITIATE_MESSAGE = "failed_to_process_qr_code_initiate_message"


class DataAgreementQrCodeProblemReport(AgentMessage):
    """Base class representing a data agreement qr code problem report message."""

    class Meta:
        """Data agreement qr code problem report metadata."""

        handler_class = DATA_AGREEMENT_QR_CODE_WORKFLOW_PROBLEM_REPORT_HANDLER_CLASS
        message_type = DATA_AGREEMENT_QR_CODE_WORKFLOW_PROBLEM_REPORT
        schema_class = "DataAgreementQrCodeProblemReportSchema"

    def __init__(
        self,
        *,
        problem_code: str = None,
        explain: str = None,
        qr_id: str = None,
        **kwargs
    ):
        """
        Initialize a DataAgreementQrCodeProblemReport message instance.

        Args:
            explain: The localized error explanation
            problem_code: The standard error identifier
            from_did: Sender DID
            to_did: Receipient DID
            created_time: The timestamp of the message
            qr_id: Qr code identifier
        """
        super().__init__(**kwargs)
        self.explain = explain
        self.problem_code = problem_code
        self.qr_id = qr_id


class DataAgreementQrCodeProblemReportSchema(AgentMessageSchema):
    """
    Data agreement qr code problem report schema.
    """
    class Meta:
        """Metadata for data agreement qr code problem report schema."""

        model_class = DataAgreementQrCodeProblemReport
        unknown = EXCLUDE

    explain = fields.Str(
        required=False,
        description="Localized error explanation",
        example="Invitation not accepted",
    )
    problem_code = fields.Str(
        data_key="problem-code",
        required=False,
        description="Standard error identifier",
        validate=validate.OneOf(
            choices=[
                dapr.value for dapr in DataAgreementQrCodeProblemReportReason],
            error="Value {input} must be one of {choices}.",
        ),
        example=DataAgreementQrCodeProblemReportReason.QR_CODE_SCANNED_ONCE.value,
    )
    qr_id = fields.Str(
        data_key="qr-id",
        required=False,
        description="Qr code identifier",
        example=UUIDFour.EXAMPLE,
    )



from aries_cloudagent.messaging.base_handler import BaseHandler, BaseResponder, RequestContext
from ..messages.data_controller_details_response import DataControllerDetailsResponseMessage

import json


class DataControllerDetailsResponseHandler(BaseHandler):
    """Handle for data-controller/1.0/details-response message"""

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """
        Message handler logic for data-controller/1.0/details-response message.
        """

        # Assert if received message is of type DataControllerDetailsResponseMessage
        assert isinstance(
            context.message, DataControllerDetailsResponseMessage)

        self._logger.info(
            "Received data-controller/1.0/details-response message: \n%s",
            json.dumps(context.message.serialize(), indent=4)
        )



from marshmallow import fields, validate, validates
from marshmallow.exceptions import ValidationError

from aries_cloudagent.messaging.models.openapi import OpenAPISchema
from aries_cloudagent.messaging.valid import (
    INDY_DID,
    UUIDFour,
    UUID4,
    INDY_RAW_PUBLIC_KEY,
)
from aries_cloudagent.connections.models.connection_record import (
    ConnectionRecord,
    ConnectionRecordSchema,
)
from aries_cloudagent.protocols.connections.v1_0.messages.connection_invitation import (
    ConnectionInvitationSchema,
)

from ...models.exchange_records.data_agreement_didcomm_transaction_record import (
    DataAgreementCRUDDIDCommTransaction,
)
from ...models.exchange_records.data_agreement_record import (
    DataAgreementV1Record,
)
from ...models.diddoc_model import MyDataDIDDocSchema

from ...utils.regex import MYDATA_DID


class ReadDataAgreementRequestSchema(OpenAPISchema):
    data_agreement_id = fields.Str(
        description="Data agreement identifier",
        required=True,
        example=UUIDFour.EXAMPLE,
    )
    connection_id = fields.UUID(
        description="Connection identifier",
        required=True,
        example=UUIDFour.EXAMPLE,
    )


class DataAgreementCRUDDIDCommTransactionResponseSchema(OpenAPISchema):
    """
    Schema for the data agreement CRUD DID comm transaction response
    """

    # Transaction identifier
    da_crud_didcomm_tx_id = fields.Str(
        example=UUIDFour.EXAMPLE,
        description="Data agreement CRUD DIDComm transaction identifier",
        required=False,
    )

    # Thread identifier
    thread_id = fields.Str(
        example=UUIDFour.EXAMPLE,
        description="Thread identifier",
        required=False,
    )

    # Message type
    message_type = fields.Str(
        example="read-data-agreement",
        description="Message type",
        required=False,
    )

    # Message list
    messages_list = fields.List(
        fields.Str(),
        description="List of messages",
        required=False,
    )

    # Connection identifier
    connection_id = fields.Str(
        example=UUIDFour.EXAMPLE,
        description="Connection identifier",
        required=False,
    )


class DACRUDDIDCommTransactionRecordListQueryStringSchema(OpenAPISchema):
    """
    Query string schema for listing data agreement CRUD DIDComm transaction records.
    """

    # Connection identifier
    connection_id = fields.UUID(
        description="Connection identifier",
        required=False,
        example=UUIDFour.EXAMPLE,
    )

    # Thread identifier
    thread_id = fields.UUID(
        description="Thread identifier",
        required=False,
        example=UUIDFour.EXAMPLE,
    )

    # Message type
    message_type = fields.Str(
        description="Message type",
        required=False,
        validate=validate.OneOf(
            [
                getattr(DataAgreementCRUDDIDCommTransaction, m)
                for m in vars(DataAgreementCRUDDIDCommTransaction)
                if m.startswith("MESSAGE_TYPE_")
            ]
        ),
    )


class DataAgreementCRUDDIDCommTransactionRecordDeleteByIdMatchInfoSchema(OpenAPISchema):
    """Delete a transaction record by its identifier."""

    da_crud_didcomm_tx_id = fields.Str(
        description="Data agreement CRUD didcomm transaction identifier",
        required=True,
        **UUID4,
    )


class CreateOrUpdateDataAgreementPersonalDataRestrictionSchema(OpenAPISchema):
    """
    Schema for the create or update data agreement personal data restriction
    """

    schema_id = fields.Str(
        description="Schema identifier",
        example="WgWxqztrNooG92RXvxSTWv:2:schema_name:1.0",
    )

    cred_def_id = fields.Str(
        description="Credential definition identifier",
        example="WgWxqztrNooG92RXvxSTWv:3:CL:20:tag",
    )


class CreateOrUpdateDataAgreementPersonalDataRequestSchema(OpenAPISchema):
    """
    Schema for the create or update data agreement personal data request
    """

    attribute_id = fields.Str(
        example=UUIDFour.EXAMPLE,
        description="Attribute identifier",
        required=False,
    )

    restrictions = fields.List(
        fields.Nested(CreateOrUpdateDataAgreementPersonalDataRestrictionSchema),
        description="List of restrictions",
        required=False,
    )


class DPIAOpenAPISchema(OpenAPISchema):
    dpia_date = fields.Str(data_key="dpiaDate")
    dpia_summary_url = fields.Str(data_key="dpiaSummaryUrl")


class PersonalDataRestrictionOpenAPISchema(OpenAPISchema):
    schema_id = fields.Str(data_key="schemaId")
    cred_def_id = fields.Str(data_key="credDefId")


class PersonalDataOpenAPISchema(OpenAPISchema):
    attribute_id = fields.Str(data_key="attributeId")
    attribute_name = fields.Str(data_key="attributeName")
    attribute_sensitive = fields.Bool(data_key="attributeSensitive")
    attribute_category = fields.Str(data_key="attributeCategory")
    attribute_description = fields.Str(data_key="attributeDescription")
    restrictions = fields.List(
        fields.Nested(PersonalDataRestrictionOpenAPISchema),
        data_key="restrictions"
    )


class DataPolicyOpenAPISchema(OpenAPISchema):
    policy_url = fields.Str(data_key="policyUrl")
    jurisdiction = fields.Str(data_key="jurisdiction")
    industry_sector = fields.Str(data_key="industrySector")
    data_retention_period = fields.Int(data_key="dataRetentionPeriod")
    geographic_restriction = fields.Str(data_key="geographicRestriction")
    storage_location = fields.Str(data_key="storageLocation")
    third_party_data_sharing = fields.Bool(data_key="thirdPartyDataSharing")


class CreateOrUpdateDataAgreementInWalletRequestSchema(OpenAPISchema):
    language = fields.Str(data_key="language")
    data_controller_name = fields.Str(data_key="dataControllerName")
    data_controller_url = fields.Str(data_key="dataControllerUrl")
    data_policy = fields.Nested(DataPolicyOpenAPISchema, data_key="dataPolicy")
    purpose = fields.Str(data_key="purpose")
    purpose_descripton = fields.Str(data_key="purposeDescription")
    lawful_basis = fields.Str(data_key="lawfulBasis")
    method_of_use = fields.Str(data_key="methodOfUse")
    personal_data = fields.List(fields.Nested(PersonalDataOpenAPISchema), data_key="personalData")
    dpia = fields.Nested(DPIAOpenAPISchema, data_key="dpia")


class CreateOrUpdateDataAgreementPersonalDataWithoutAttributeIdSchema(OpenAPISchema):
    """
    Personal data schema class
    """

    @validates("attribute_name")
    def validate_attribute_name(self, attribute_name):
        """
        Validate attribute name
        """
        if len(attribute_name) < 3:
            raise ValidationError("Attribute name must be at least 3 characters long")

    # Attribute name
    attribute_name = fields.Str(
        example="Name", description="Name of the attribute", required=True
    )

    # Attribute description
    attribute_description = fields.Str(
        required=True,
        description="The description of the attribute.",
        example="Name of the customer",
    )

    restrictions = fields.List(
        fields.Nested(CreateOrUpdateDataAgreementPersonalDataRestrictionSchema),
        description="List of restrictions",
        required=False,
    )


class CreateOrUpdateDataAgreementInWalletRequestSchemaV2(
    CreateOrUpdateDataAgreementInWalletRequestSchema
):
    # Data agreement personal data (attributes)
    personal_data = fields.List(
        fields.Nested(CreateOrUpdateDataAgreementPersonalDataWithoutAttributeIdSchema),
        required=True,
    )


class DataAgreementV1RecordResponseSchema(OpenAPISchema):
    """
    Schema for data agreement v1 record response
    """
    template_id = fields.Str()
    state = fields.Str()
    method_of_use = fields.Str()
    data_agreement = fields.Dict()
    schema_id = fields.Str()
    cred_def_id = fields.Str()
    presentation_request = fields.Dict()
    publish_flag = fields.Str()
    delete_flag = fields.Str()
    existing_schema_flag = fields.Str()
    latest_version_flag = fields.Str()


class AuditorDIDCommTransactionRecordsRetrieveByIdMatchInfoSchema(OpenAPISchema):
    """
    Retrieve a transaction record by its identifier.
    """

    auditor_didcomm_transaction_record_id = fields.Str(
        description="Auditor didcomm transaction identifier", required=True, **UUID4
    )


class AuditorDIDCommTransactionRecordsDeleteByIdMatchInfoSchema(OpenAPISchema):
    """
    Delete a transaction record by its identifier.
    """

    auditor_didcomm_transaction_record_id = fields.Str(
        description="Auditor didcomm transaction identifier", required=True, **UUID4
    )


class AuditorDIDCommTransactionRecordListQueryStringSchema(OpenAPISchema):
    """
    Query string schema for listing Auditor DIDComm transaction records.
    """

    # Connection identifier
    connection_id = fields.UUID(
        description="Connection identifier",
        required=False,
        example=UUIDFour.EXAMPLE,
    )

    # Thread identifier
    thread_id = fields.UUID(
        description="Thread identifier",
        required=False,
        example=UUIDFour.EXAMPLE,
    )


class DataAgreementQueryStringSchema(OpenAPISchema):
    """
    Query string schema for data agreements
    """

    method_of_use = fields.Str(
        description="Method of use",
        required=False,
        validate=validate.OneOf(
            [
                getattr(DataAgreementV1Record, m)
                for m in vars(DataAgreementV1Record)
                if m.startswith("METHOD_OF_USE_")
            ]
        ),
    )

    template_id = fields.Str(
        description="Template identifier",
        required=False,
    )

    template_version = fields.Str(
        description="Template version",
    )

    delete_flag = fields.Bool(
        description="Query deleted templates",
        required=False,
    )

    publish_flag = fields.Bool(
        description="Query published templates",
        required=False,
    )

    third_party_data_sharing = fields.Bool(
        description="Third party data sharing",
        required=False,
    )

    latest_version_flag = fields.Bool(
        description="Latest version of the template",
        required=False,
    )

    page = fields.Int(
        required=False,
        description="Page number",
        example=1,
    )

    page_size = fields.Int(
        required=False,
        description="Page size",
        example=10,
    )


class UpdateDataAgreementMatchInfoSchema(OpenAPISchema):
    """
    Schema to match info for the update data agreement endpoint
    """

    template_id = fields.Str(
        description="Template identifier", required=True
    )


class DeleteDataAgreementMatchInfoSchema(OpenAPISchema):
    """
    Schema to match info for the delete data agreement endpoint
    """

    template_id = fields.Str(
        description="Template identifier", required=True
    )


class CreateAndStoreDAPersonalDataInWalletRequestSchema(OpenAPISchema):

    """
    Schema for the create and store personal data in wallet request
    """

    @validates("attribute_name")
    def validate_attribute_name(self, attribute_name):
        """
        Validate attribute name
        """
        if len(attribute_name) < 3:
            raise ValidationError("Attribute name must be at least 3 characters long")

    # Attribute name
    attribute_name = fields.Str(
        example="Name", description="Name of the attribute", required=True
    )

    # Attribute sensitive
    attribute_sensitive = fields.Bool(
        example=True, description="Sensitivity of the attribute", required=False
    )

    # Attribute category
    attribute_category = fields.Str(
        example="Personal", description="Category of the attribute", required=False
    )

    @validates("attribute_description")
    def validate_attribute_description(self, attribute_description):
        """
        Validate attribute description
        """
        if len(attribute_description) < 3:
            raise ValidationError(
                "Attribute description must be at least 3 characters long"
            )

        if len(attribute_description) > 1000:
            raise ValidationError(
                "Attribute description must be at most 1000 characters long"
            )

    # Attribute description
    attribute_description = fields.Str(
        example="Name of the user",
        description="Description of the attribute",
        required=False,
    )


class QueryDAPersonalDataInWalletQueryStringSchema(OpenAPISchema):
    """
    Schema for the query personal data in wallet query string
    """

    attribute_sensitive = fields.Bool(
        description="Sensitivity of the attribute", required=False
    )

    attribute_category = fields.Str(
        description="Category of the attribute", required=False
    )


class ListDAPersonalDataCategoryFromWalletResponseSchema(OpenAPISchema):
    """
    Schema for the list personal data category from wallet response
    """

    # List of categories
    categories = fields.List(
        fields.Str(description="Category", example="Personal"),
        description="List of categories",
        required=True,
    )


class MarkExistingConnectionAsAuditorMatchInfoSchema(OpenAPISchema):
    """
    Schema for the mark existing connection as auditor match info
    """

    connection_id = fields.Str(
        example=UUIDFour.EXAMPLE, description="Connection identifier", required=True
    )


class QueryDataAgreementInstanceQueryStringSchema(OpenAPISchema):
    """
    Query data agreement instances
    """

    instance_id = fields.Str(required=False)
    template_id = fields.Str(required=False)
    template_version = fields.Str(required=False)
    method_of_use = fields.Str(
        required=False,
        validate=validate.OneOf(
            [
                "data-source",
                "data-using-service",
            ]
        ),
    )
    third_party_data_sharing = fields.Bool(required=False)
    data_ex_id = fields.Str(required=False)
    data_subject_did = fields.Str(required=False)
    page = fields.Int(required=False)
    page_size = fields.Int(required=False)


class AuditorSendDataAgreementVerifyRequestMatchInfoSchema(OpenAPISchema):
    """
    Schema to send data agreement verify request to the auditor
    """

    data_agreement_id = fields.Str(
        description="Data agreement identifier",
        required=True,
        example=UUIDFour.EXAMPLE,
    )


class DataAgreementQRCodeMatchInfoSchema(OpenAPISchema):
    """
    Schema for data agreement QR code match info
    """

    template_id = fields.Str(required=True)


class DataAgreementQRCodeInvitationSchema(OpenAPISchema):
    """Schema for connection invitation details inside in data agreement qr code payload."""

    service_endpoint = fields.Str(
        description="Service endpoint", example="http://localhost:8080/"
    )
    recipient_key = fields.Str(description="Recipient key", **INDY_RAW_PUBLIC_KEY)


class GenerateDataAgreementQrCodePayloadResponseSchema(OpenAPISchema):
    """
    Schema for Data Agreement QR code payload
    """

    qr_id = fields.Str(description="QR code ID", **UUID4)
    connection_id = fields.Str(description="Connection ID", **UUID4)
    invitation = fields.Nested(
        DataAgreementQRCodeInvitationSchema(),
        description="Connection invitation information",
    )


class SendReadDIDMessageMatchInfoSchema(OpenAPISchema):
    """
    Send a read-did message to the MyData DID registry service.
    """

    did = fields.Str(description="did:mydata identifier", required=True, **MYDATA_DID)


class MyDataDIDRemoteRecordsQueryStringSchema(OpenAPISchema):
    """
    Query string schema for listing MyData DID remote records.
    """

    # Sovrin verkey
    sov_verkey = fields.Str(
        description="Sovrin verkey", required=False, **INDY_RAW_PUBLIC_KEY
    )

    # DID
    did = fields.Str(**MYDATA_DID, description="MyData decentralised identifier")

    # Status
    status = fields.Str(
        description="MyData DID remote status",
        required=False,
        validate=validate.OneOf(["active", "revoked"]),
    )


class MyDataDIDRemoteRecordResponseSchema(OpenAPISchema):
    """
    Response schema for MyData DID remote record.
    """

    did_doc = fields.Nested(
        MyDataDIDDocSchema,
        description="MyData DID document",
    )

    did = fields.Str(**MYDATA_DID, description="MyData decentralised identifier")

    sov_verkey = fields.Str(
        description="Sovrin verkey", required=False, **INDY_RAW_PUBLIC_KEY
    )

    status = fields.Str(
        description="MyData DID remote status",
        required=False,
        validate=validate.OneOf(["active", "revoked"]),
    )


class CreateOrUpdateDataAgreementInWalletQueryStringSchema(OpenAPISchema):
    """Query string schema for create data agreement handler"""

    publish_flag = fields.Boolean(description="Publish the agreement",
                                  required=False, example=False)

    existing_schema_id = fields.Str(
        description="Existing schema identifier",
        required=False,
        example="issuer_did:1:schema:1",
    )


class UpdateDataAgreementTemplateOpenAPISchema(OpenAPISchema):
    publish_flag = fields.Boolean(required=False)
    existing_schema_id = fields.Str(required=False)


class PublishDataAgreementMatchInfoSchema(OpenAPISchema):
    """
    Schema to match info for the publish data agreement endpoint
    """

    template_id = fields.Str()


class QueryDaPersonalDataInWalletQueryStringSchema(OpenAPISchema):
    template_id = fields.Str(required=False)
    page = fields.Int(required=False)
    page_size = fields.Int(required=False)
    method_of_use = fields.Str(
        required=False,
        validate=validate.OneOf(
            [
                "data-source",
                "data-using-service",
            ]
        ),
    )
    third_party_data_sharing = fields.Bool(required=False)


class UpdateDaPersonalDataInWalletMatchInfoSchema(OpenAPISchema):

    attribute_id = fields.Str(
        required=True, description="Personal data identifier", example=UUIDFour.EXAMPLE
    )


class UpdateDaPersonalDataInWalletRequestSchema(OpenAPISchema):
    attribute_description = fields.Str(
        description="Attribute description", example="Age of the patient", required=True
    )


class UpdateDaPersonalDataInWalletResponseSchema(OpenAPISchema):

    attribute_id = fields.Str(description="Attribute ID", example=UUIDFour.EXAMPLE)
    attribute_name = fields.Str(description="Attribute name", example="Name")
    attribute_description = fields.Str(
        description="Attribute description", example="Name of the patient"
    )
    data_agreement_template_id = fields.Str(
        description="Data Agreement Template ID", example=UUIDFour.EXAMPLE
    )
    data_agreement_template_version = fields.Integer(
        description="Data Agreement Template version", example=1
    )
    created_at = fields.Integer(
        description="Created at (Epoch time in seconds)", example=1578012800
    )
    updated_at = fields.Integer(
        description="Updated at (Epoch time in seconds)", example=1578012800
    )


class DeleteDaPersonalDataInWalletMatchInfoSchema(OpenAPISchema):

    attribute_id = fields.Str(
        required=True, description="Personal data identifier", example=UUIDFour.EXAMPLE
    )


class QueryDataAgreementQrCodeMetadataRecordsMatchInfoSchema(OpenAPISchema):
    """Schema to validate path parameters for query data agreement qr code metadata records."""

    template_id = fields.Str(
        description="Template identifier",
        example=UUIDFour.EXAMPLE,
        required=False,
    )


class QueryDataAgreementQrCodeMetadataRecordsQueryStringSchema(OpenAPISchema):
    """Schema for querying data agreement qr code metadata records"""

    qr_id = fields.Str(
        description="QR code identifier", example=UUIDFour.EXAMPLE, required=False
    )


class QueryDataAgreementQRCodeMetadataRecordsResponseSchema(OpenAPISchema):
    """Schema for querying data agreement qr code metadata records response"""

    qr_id = fields.Str(
        description="QR code identifier", example=UUIDFour.EXAMPLE, required=False
    )

    data_agreement_id = fields.Str(
        description="Data Agreement identifier",
        example=UUIDFour.EXAMPLE,
        required=False,
    )

    connection_id = fields.Str(
        description="Connection identifier", example=UUIDFour.EXAMPLE, required=False
    )

    multi_use = fields.Bool()
    is_scanned = fields.Bool()
    data_exchange_record_id = fields.Str()


class RemoveDataAgreementQrCodeMetadataRecordMatchInfoSchema(OpenAPISchema):
    template_id = fields.Str()
    qr_id = fields.Str()


class Base64EncodeDataAgreementQrCodeMatchInfoSchema(OpenAPISchema):
    """Match info (URL path params) schema for base64 encode
    data agreement QR code payload endpoint"""

    # Data Agreement identifier.
    data_agreement_id = fields.Str(
        description="Data Agreement identifier", example=UUIDFour.EXAMPLE, required=True
    )

    # Qr code identifier
    qr_id = fields.Str(
        description="QR code identifier", example=UUIDFour.EXAMPLE, required=True
    )


class SendDataAgreementQrCodeWorkflowInitiateHandlerMatchInfoSchema(OpenAPISchema):
    """Match info (URL path params) schema for send
    data agreement qr code workflow initiate endpoint"""

    # Connection identifier.
    connection_id = fields.Str(
        description="Connection identifier", example=UUIDFour.EXAMPLE, required=True
    )

    # Qr code identifier
    qr_id = fields.Str(
        description="QR code identifier", example=UUIDFour.EXAMPLE, required=True
    )


class SendReadAllDataAgreementTemplateMessageHandlerMatchInfoSchema(OpenAPISchema):
    """Schema for matching path parameters in send read all
    data agreement template message handler"""

    connection_id = fields.Str(
        description="Connection identifier", example=UUIDFour.EXAMPLE, required=True
    )


class GenerateDataAgreementQrCodePayloadQueryStringSchema(OpenAPISchema):
    """Schema for query string parameters to generate data agreement qr code payload"""

    multi_use = fields.Bool()


class SendJSONLDDIDCommProcessedDataMessageHandlerRequestSchema(OpenAPISchema):
    data = fields.Dict()
    signature_options = fields.Dict()
    proof_chain = fields.Bool()


class SendJSONLDDIDCommProcessedDataMessageHandlerMatchInfoSchema(OpenAPISchema):

    connection_id = fields.Str(
        description="Connection identifier", example=UUIDFour.EXAMPLE, required=True
    )


class V2CreateInvitationQueryStringSchema(OpenAPISchema):
    """Parameters and validators for create invitation request query string."""

    alias = fields.Str(
        description="Alias",
        required=False,
        example="Barry",
    )
    auto_accept = fields.Boolean(
        description="Auto-accept connection (default as per configuration)",
        required=False,
    )
    public = fields.Boolean(
        description="Create invitation from public DID (default false)", required=False
    )
    multi_use = fields.Boolean(
        description="Create invitation for multiple use (default false)", required=False
    )


class V2InvitationResultSchema(OpenAPISchema):
    """Result schema for a new connection invitation."""

    connection_id = fields.Str(
        description="Connection identifier", example=UUIDFour.EXAMPLE
    )
    invitation = fields.Nested(ConnectionInvitationSchema())
    invitation_url = fields.Str(
        description="Invitation URL",
        example="http://192.168.56.101:8020/invite?c_i=eyJAdHlwZSI6Li4ufQ==",
    )


class GenerateFirebaseDynamicLinkForConnectionInvitationMatchInfoSchema(OpenAPISchema):
    """Schema for matching path parameters in generate firebase
    dynamic link for connection invitation handler"""

    conn_id = fields.Str(
        description="Connection identifier", example=UUIDFour.EXAMPLE, required=True
    )


class GenerateFirebaseDynamicLinkForConnectionInvitationResponseSchema(OpenAPISchema):
    """Schema for response of generate firebase dynamic link for connection invitation handler"""

    # Firebase dynamic link
    firebase_dynamic_link = fields.Str(
        description="Firebase dynamic link",
        example="https://example.page.link/UVWXYZuvwxyz12345",
    )


class SendExistingConnectionsMessageHandlerMatchInfoSchema(OpenAPISchema):
    """Schema for matching path parameters in send existing connections message handler"""

    conn_id = fields.Str(
        description="Connection identifier", example=UUIDFour.EXAMPLE, required=True
    )


class SendExistingConnectionsMessageHandlerRequestSchema(OpenAPISchema):
    """Schema for request body of send existing connections message handler"""

    theirdid = fields.Str(
        description="Their DID", example="QmWbsNYhMrjHiqZDTUASHg", required=True
    )


class GetExistingConnectionMatchInfoSchema(OpenAPISchema):
    """Schema for matching path parameters in get existing connection handler"""

    conn_id = fields.Str(
        description="Connection identifier", example=UUIDFour.EXAMPLE, required=True
    )


class GetExistingConnectionResponseSchema(OpenAPISchema):
    """Schema for response of get existing connection handler"""

    existing_connection_id = fields.Str()
    my_did = fields.Str()
    connection_status = fields.Str()
    connection_id = fields.Str()


class ConnectionsListQueryStringSchemaV2(OpenAPISchema):
    """Parameters and validators for connections list request query string."""

    alias = fields.Str(
        description="Alias",
        required=False,
        example="Barry",
    )

    initiator = fields.Str(
        description="Connection initiator",
        required=False,
        validate=validate.OneOf(["self", "external"]),
    )

    invitation_key = fields.Str(
        description="invitation key", required=False, **INDY_RAW_PUBLIC_KEY
    )

    my_did = fields.Str(description="My DID", required=False, **INDY_DID)

    state = fields.Str(
        description="Connection state",
        required=False,
        validate=validate.OneOf(
            [
                getattr(ConnectionRecord, m)
                for m in vars(ConnectionRecord)
                if m.startswith("STATE_")
            ]
        ),
    )

    their_did = fields.Str(description="Their DID", required=False, **INDY_DID)

    their_role = fields.Str(
        description="Their assigned connection role",
        required=False,
        example="Point of contact",
    )

    # Page
    page = fields.Int(
        required=False,
        description="Page number",
        example=1,
    )

    # Page size
    page_size = fields.Int(
        required=False,
        description="Page size",
        example=10,
    )

    # Org flag
    org_flag = fields.Bool(required=False)

    # Marketplace flag
    marketplace_flag = fields.Bool(required=False)


class ConnectionListSchema(OpenAPISchema):
    """Result schema for connection list."""

    results = fields.List(
        fields.Nested(ConnectionRecordSchema()),
        description="List of connection records",
    )


class UpdateControllerDetailsRequestSchema(OpenAPISchema):
    """Update controller details request schema"""
    organisation_did = fields.Str(required=False)
    organisation_name = fields.Str(required=False)
    cover_image_url = fields.Str(required=False)
    logo_image_url = fields.Str(required=False)
    location = fields.Str(required=False)
    organisation_type = fields.Str(required=False)
    description = fields.Str(required=False)
    policy_url = fields.Str(required=False)
    eula_url = fields.Str(required=False)


class ConfigureCustomerIdentificationDAMatchInfoSchema(OpenAPISchema):
    """Configure customer identification DA match info schema"""

    template_id = fields.Str()



from enum import Enum
from marshmallow import EXCLUDE, fields, validate

from aries_cloudagent.messaging.agent_message import AgentMessage, AgentMessageSchema
from aries_cloudagent.messaging.valid import UUIDFour

from ..message_types import (
    PROTOCOL_PACKAGE,
    JSON_LD_PROBLEM_REPORT
)
from ..utils.regex import MYDATA_DID

# Handler class path for JSONLD Problem Report (json-ld/1.0/problem-report) message
JSON_LD_PROBLEM_REPORT_HANDLER_CLASS = (
    f"{PROTOCOL_PACKAGE}.handlers"
    ".json_ld_problem_report_handler.JSONLDProblemReportHandler"
)


class JSONLDProblemReportReason(str, Enum):
    """Supported reason codes."""

    # Failed to generate processed data due to invalid input
    INVALID_INPUT = "invalid_input"


class JSONLDProblemReport(AgentMessage):
    """Base class representing a JSON-LD problem report message."""

    class Meta:
        """JSON-LD problem report metadata."""

        handler_class = JSON_LD_PROBLEM_REPORT_HANDLER_CLASS
        message_type = JSON_LD_PROBLEM_REPORT
        schema_class = "JSONLDProblemReportSchema"

    def __init__(
        self,
        *,
        problem_code: str = None,
        explain: str = None,
        from_did: str = None,
        to_did: str = None,
        created_time: str = None,
        **kwargs
    ):
        """
        Initialize a JSONLDProblemReport message instance.

        Args:
            explain: The localized error explanation
            problem_code: The standard error identifier
            from_did: Sender DID
            to_did: Receipient DID
            created_time: The timestamp of the message
        """
        super().__init__(**kwargs)
        self.explain = explain
        self.problem_code = problem_code
        self.from_did = from_did
        self.to_did = to_did
        self.created_time = created_time


class JSONLDProblemReportSchema(AgentMessageSchema):
    """
    JSON-LD problem report schema.
    """
    class Meta:
        """Metadata for JSON-LD problem report schema."""

        model_class = JSONLDProblemReport
        unknown = EXCLUDE

    from_did = fields.Str(data_key="from", **MYDATA_DID)
    to_did = fields.Str(data_key="to", **MYDATA_DID)
    created_time = fields.Str(data_key="created_time")
    explain = fields.Str(
        required=False,
        description="Localized error explanation",
        example="Invitation not accepted",
    )
    problem_code = fields.Str(
        data_key="problem-code",
        required=False,
        description="Standard error identifier",
        validate=validate.OneOf(
            choices=[
                dapr.value for dapr in JSONLDProblemReportReason],
            error="Value {input} must be one of {choices}.",
        ),
        example=JSONLDProblemReportReason.INVALID_INPUT.value,
    )


"""Credential offer message handler."""
from aries_cloudagent.messaging.base_handler import (
    BaseHandler,
    BaseResponder,
    HandlerException,
    RequestContext,
)
from dexa_sdk.managers.ada_manager import V2ADAManager
from ..manager import CredentialManager
from ..messages.credential_offer import CredentialOffer
from .....v1_0.decorators.data_agreement_context_decorator import DataAgreementContextDecorator
from aries_cloudagent.utils.tracing import trace_event, get_timer


class CredentialOfferHandler(BaseHandler):
    """Message handler class for credential offers."""

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """
        Message handler logic for credential offers.

        Args:
            context: request context
            responder: responder callback

        """
        r_time = get_timer()

        self._logger.debug(
            "CredentialOfferHandler called with context %s", context)
        assert isinstance(context.message, CredentialOffer)
        self._logger.info(
            "Received credential offer message: %s",
            context.message.serialize(as_string=True),
        )

        if not context.connection_ready:
            raise HandlerException(
                "No connection established for credential offer")

        credential_manager = CredentialManager(context)

        cred_ex_record = await credential_manager.receive_offer()

        # Initialise ADA manager
        manager = V2ADAManager(context)

        # Process data agreement context decorator if present.
        # Create data agreement instance from da offer.
        instance_record = await manager.process_decorator_with_da_offer_message(
            context.message._decorators,
            cred_ex_record,
            context.connection_record
        )

        r_time = trace_event(
            context.settings,
            context.message,
            outcome="CredentialOfferHandler.handle.END",
            perf_counter=r_time,
        )

        # If auto respond is turned on, automatically reply with credential request
        if context.settings.get("debug.auto_respond_credential_offer"):
            (_, credential_request_message) = await credential_manager.create_request(
                cred_ex_record=cred_ex_record,
                holder_did=context.connection_record.my_did,
            )

            # Build data agreement negotiation accept message
            accept_message = \
                await manager.build_data_agreement_negotiation_accept_by_instance_id(
                    instance_record.instance_id,
                    context.connection_record
                )

            # Update credential request message with data agreement context decorator
            credential_request_message._decorators["data-agreement-context"] = \
                DataAgreementContextDecorator(
                message_type="protocol",
                message=accept_message.serialize()
            )

            await responder.send_reply(credential_request_message)

            trace_event(
                context.settings,
                credential_request_message,
                outcome="CredentialOfferHandler.handle.REQUEST",
                perf_counter=r_time,
            )



import os
import typing
import logging
import jwt

from aiohttp import web
from aiohttp import frozenlist
from ..manager import ADAManager, ADAManagerError

from .maps.tag_maps import (
    TAGS_MYDATA_DID_OPERATIONS,
    TAGS_DATA_AGREEMENT_CORE_FUNCTIONS,
    TAGS_DATA_AGREEMENT_AUDITOR_FUNCTIONS,
    TAGS_JSONLD_FUNCTIONS,
    TAGS_DATA_CONTROLLER_FUNCTIONS
)

from .maps.route_maps import ROUTES_ADA


LOGGER = logging.getLogger(__name__)

PAGINATION_PAGE_SIZE = 10


@web.middleware
async def authentication_middleware(
    request: web.BaseRequest, handler: typing.Coroutine
):
    """
    Authentication middleware.

    Authenticate the request if the request headers contains
    Authorization header with value of ApiKey <api_key>.
    """

    # Context.
    context = request.app["request_context"]

    # Fetch iGrant.io config from os environment.
    config_org_id = context.settings.get("intermediary.igrantio_org_id")
    config_api_key_secret = context.settings.get("intermediary.igrantio_org_api_key_secret")

    if not config_org_id or "webhook" in request.path:
        # Intermediary config not available.
        return await handler(request)

    # Fetch authorization header.
    authorization_header = request.headers.get("Authorization")

    # Fetch api key from authorization header.
    header_api_key = authorization_header.split("ApiKey ")[1] if authorization_header else None

    if not header_api_key:
        raise web.HTTPUnauthorized(reason="Missing Authorization header.")

    # Authenticate the request.
    try:
        jwt.decode(header_api_key, config_api_key_secret, algorithms=["HS256"])
    except jwt.exceptions.InvalidTokenError:
        try:
            jwt.decode(
                header_api_key, config_api_key_secret, algorithms=["HS256"], audience="dataverifier"
            )
        except jwt.exceptions.InvalidTokenError:
            raise web.HTTPUnauthorized(reason="Invalid API Key.")

    # Override settings api key.
    context.settings["intermediary.igrantio_org_api_key"] = header_api_key

    # Call the handler.
    return await handler(request)


async def register(app: web.Application):

    app._middlewares = frozenlist.FrozenList(
        app.middlewares[:] + [authentication_middleware]
    )

    app.add_routes(ROUTES_ADA)


def post_process_routes(app: web.Application):
    """Amend swagger API."""

    # Add top-level tags description
    if "tags" not in app._state["swagger_dict"]:
        app._state["swagger_dict"]["tags"] = []

    app._state["swagger_dict"]["tags"].append(TAGS_MYDATA_DID_OPERATIONS)
    app._state["swagger_dict"]["tags"].append(TAGS_DATA_AGREEMENT_CORE_FUNCTIONS)
    app._state["swagger_dict"]["tags"].append(TAGS_DATA_AGREEMENT_AUDITOR_FUNCTIONS)
    app._state["swagger_dict"]["tags"].append(TAGS_JSONLD_FUNCTIONS)
    app._state["swagger_dict"]["tags"].append(TAGS_DATA_CONTROLLER_FUNCTIONS)



import datetime

from aries_cloudagent.messaging.models.base import BaseModel, BaseModelSchema
from aries_cloudagent.messaging.valid import UUIDFour
from marshmallow import fields, EXCLUDE
from marshmallow.exceptions import ValidationError
from typing import List

from ..utils.regex import MYDATA_DID
from .data_agreement_instance_model import DataAgreementInstanceSchema, DataAgreementInstance

class ReadDataAgreementResponseBody(BaseModel):
    """ReadDataAgreementResponseBody model class"""

    class Meta:
        """ReadDataAgreementResponseBody metadata"""
        schema_class = "ReadDataAgreementResponseBodySchema"

    def __init__(
        self,
        *,
        data_agreement :DataAgreementInstance = None,
        **kwargs,
    ):
        """
        Initialize ReadDataAgreementResponseBody model

        :param data_agreement: Data Agreement
        :param kwargs: kwargs
        """
        super().__init__(**kwargs)
        self.data_agreement = data_agreement


class ReadDataAgreementResponseBodySchema(BaseModelSchema):
    """ReadDataAgreementResponseBody schema class"""

    class Meta:
        """ReadDataAgreementResponseBodySchema metadata"""
        model_class = ReadDataAgreementResponseBody
        unknown = EXCLUDE

    data_agreement = fields.Nested(DataAgreementInstanceSchema())


import logging
from aiohttp import web
from aiohttp_apispec import (
    docs,
    querystring_schema,
    response_schema,
)

from aries_cloudagent.messaging.models.base import BaseModelError
from aries_cloudagent.storage.error import StorageError
from dexa_sdk.managers.ada_manager import V2ADAManager
from dexa_sdk.utils import clean_and_get_field_from_dict
from ..models.data_agreement_instance_model import (
    DataAgreementInstanceSchema,
)
from ..routes.maps.tag_maps import (
    TAGS_DATA_AGREEMENT_AUDITOR_FUNCTIONS_LABEL,
)

from .openapi import (
    QueryDataAgreementInstanceQueryStringSchema,
)

LOGGER = logging.getLogger(__name__)

PAGINATION_PAGE_SIZE = 10


@docs(
    tags=[TAGS_DATA_AGREEMENT_AUDITOR_FUNCTIONS_LABEL], summary="Query data agreement instances"
)
@querystring_schema(QueryDataAgreementInstanceQueryStringSchema())
@response_schema(DataAgreementInstanceSchema(many=True), 200)
async def query_data_agreement_instances(request: web.BaseRequest):
    """
    Request handler for querying data agreement instances.
    """

    # Context
    context = request.app["request_context"]

    instance_id = clean_and_get_field_from_dict(request.query, "instance_id")
    template_id = clean_and_get_field_from_dict(request.query, "template_id")
    template_version = clean_and_get_field_from_dict(request.query, "template_version")
    method_of_use = clean_and_get_field_from_dict(request.query, "method_of_use")
    third_party_data_sharing = clean_and_get_field_from_dict(
        request.query, "third_party_data_sharing")
    data_ex_id = clean_and_get_field_from_dict(request.query, "data_ex_id")
    data_subject_did = clean_and_get_field_from_dict(request.query, "data_subject_did")
    page = clean_and_get_field_from_dict(request.query, "page")
    page = int(page) if page is not None else page
    page_size = clean_and_get_field_from_dict(request.query, "page_size")
    page_size = int(page_size) if page_size is not None else page_size

    try:

        # Initialise MyData DID Manager
        manager: V2ADAManager = V2ADAManager(context=context)

        # Get the data agreement instances
        paginationResult = await manager.query_data_agreement_instances(
            instance_id,
            template_id,
            template_version,
            method_of_use,
            third_party_data_sharing,
            data_ex_id,
            data_subject_did,
            page if page else 1,
            page_size if page_size else 10
        )

    except (StorageError, BaseModelError) as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response(paginationResult._asdict())



import datetime

import validators

from aries_cloudagent.messaging.models.base import BaseModel, BaseModelSchema
from aries_cloudagent.messaging.valid import UUIDFour
from marshmallow import fields, EXCLUDE, validate, validates, pre_load
from marshmallow.exceptions import ValidationError
from typing import List

from ..utils.regex import MYDATA_DID

DATA_AGREEMENT_V1_SCHEMA_CONTEXT = "https://raw.githubusercontent.com/decentralised-dataexchange/automated-data-agreements/main/interface-specs/data-agreement-schema/v1/data-agreement-schema-context.jsonld"


class DataAgreementDataPolicy(BaseModel):
    """
    Data policy model class
    """

    class Meta:
        # Schema class
        schema_class = "DataAgreementDataPolicySchema"

    def __init__(
        self,
        *,
        data_retention_period: int,
        policy_url: str,
        jurisdiction: str,
        industry_scope: str,
        geographic_restriction: str,
        storage_location: str,
        third_party_data_sharing: bool,
        **kwargs,
    ):
        """
        Initialize data policy model
        """

        # Call parent constructor
        super().__init__(**kwargs)

        # Set attributes
        self.data_retention_period = data_retention_period
        self.policy_url = policy_url
        self.jurisdiction = jurisdiction
        self.industry_scope = industry_scope
        self.geographic_restriction = geographic_restriction
        self.storage_location = storage_location
        self.third_party_data_sharing = third_party_data_sharing


class DataAgreementDataPolicySchema(BaseModelSchema):
    """
    Data policy schema class
    """

    class Meta:
        # Model class
        model_class = DataAgreementDataPolicy

        # Unknown fields are excluded
        unknown = EXCLUDE

    @validates("data_retention_period")
    def validate_data_retention_period(self, data_retention_period):
        """
        Validate data retention period
        """
        if data_retention_period < 1:
            raise ValidationError(
                "Data retention period must be greater than or equal to 1"
            )

    # Data retention period
    data_retention_period = fields.Int(
        data_key="data_retention_period",
        example=365,
        description="Data retention period in days",
        required=True,
    )

    @validates("policy_url")
    def validate_policy_url(self, policy_url):
        """
        Validate policy url
        """
        if not validators.url(policy_url):
            raise ValidationError("Policy URL is not valid")

    # Policy URL
    policy_url = fields.Str(
        data_key="policy_URL",
        example="https://clarifyhealth.com/privacy-policy/",
        description="Policy URL",
        required=True,
    )

    # Jurisdiction
    jurisdiction = fields.Str(
        data_key="jurisdiction",
        example="Sweden",
        description="Jurisdiction",
        required=True,
    )

    # Industry scope
    industry_scope = fields.Str(
        data_key="industry_sector",
        example="Healthcare",
        description="Industry scope",
        required=False,
    )

    # Geographic restriction
    geographic_restriction = fields.Str(
        data_key="geographic_restriction",
        example="Europe",
        description="Geographic restriction",
        required=True,
    )

    # Storage location
    storage_location = fields.Str(
        data_key="storage_location",
        example="Europe",
        description="Storage location",
        required=True,
    )

    third_party_data_sharing = fields.Bool(
        data_key="third_party_data_sharing",
        example=False,
        description="Third party data sharing",
        required=True,
    )


class DataAgreementDPIA(BaseModel):
    """
    DPIA model class
    """

    class Meta:
        # Schema class
        schema_class = "DataAgreementDPIASchema"

    def __init__(self, *, dpia_date: str, dpia_summary_url: str, **kwargs):
        """
        Initialize DPIA model
        """
        # Call parent constructor
        super().__init__(**kwargs)

        # Set attributes
        self.dpia_date = dpia_date
        self.dpia_summary_url = dpia_summary_url


class DataAgreementDPIASchema(BaseModelSchema):
    """
    DPIA schema class
    """

    class Meta:
        # Model class
        model_class = DataAgreementDPIA

        # Unknown fields are excluded
        unknown = EXCLUDE

    # DPIA date
    dpia_date = fields.Str(
        data_key="dpia_date",
        example=str(
            datetime.datetime.utcnow().replace(tzinfo=datetime.timezone.utc).isoformat()
        ),
        description="DPIA conducted date time in ISO 8601 UTC date time format",
        required=False,
    )

    # DPIA summary URL
    dpia_summary_url = fields.Str(
        data_key="dpia_summary_url",
        example="https://org.com/dpia_results.html",
        description="DPIA summary URL",
        required=False,
    )


class DataAgreementPersonalDataRestriction(BaseModel):
    """
    Personal data restriction model class
    """

    class Meta:
        # Schema class
        schema_class = "DataAgreementPersonalDataRestrictionSchema"

    def __init__(self, *, schema_id: str = None, cred_def_id: str = None, **kwargs):
        """
        Initialise personal data restriction model.
        """

        # Call parent constructor
        super().__init__(**kwargs)

        # Set attributes
        self.schema_id = schema_id
        self.cred_def_id = cred_def_id


class DataAgreementPersonalDataRestrictionSchema(BaseModelSchema):
    """
    Personal data restriction schema class
    """

    class Meta:
        # Model class
        model_class = DataAgreementPersonalDataRestriction

    schema_id = fields.Str(
        description="Schema identifier",
        example="WgWxqztrNooG92RXvxSTWv:2:schema_name:1.0",
        required=False,
    )

    cred_def_id = fields.Str(
        description="Credential definition identifier",
        example="WgWxqztrNooG92RXvxSTWv:3:CL:20:tag",
        required=False,
    )

    @pre_load
    def unwrap_envelope(self, data: dict, **kwargs):

        if len(data.values()) < 1:
            raise ValidationError(
                "Personal data restriction must contain at least one attribute."
            )

        return data


class DataAgreementPersonalData(BaseModel):
    """
    Personal data model class
    """

    class Meta:
        # Schema class
        schema_class = "DataAgreementPersonalDataSchema"

    def __init__(
        self,
        *,
        attribute_id: str = None,
        attribute_name: str = None,
        attribute_sensitive: bool = None,
        attribute_category: str = None,
        attribute_description: str = None,
        restrictions: List[DataAgreementPersonalDataRestriction] = None,
        **kwargs,
    ):
        """
        Initialize personal data model
        """

        # Call parent constructor
        super().__init__(**kwargs)

        # Set attributes
        self.attribute_id = attribute_id
        self.attribute_name = attribute_name
        self.attribute_sensitive = attribute_sensitive
        self.attribute_category = attribute_category
        self.attribute_description = attribute_description
        self.restrictions = restrictions


class DataAgreementPersonalDataSchema(BaseModelSchema):
    """
    Personal data schema class
    """

    class Meta:
        # Model class
        model_class = DataAgreementPersonalData

    # Attribute identifier
    attribute_id = fields.Str(
        example=UUIDFour.EXAMPLE,
        description="Attribute identifier",
        required=False,
    )

    @validates("attribute_name")
    def validate_attribute_name(self, attribute_name):
        """
        Validate attribute name
        """
        if len(attribute_name) < 3:
            raise ValidationError("Attribute name must be at least 3 characters long")

    # Attribute name
    attribute_name = fields.Str(
        example="Name", description="Name of the attribute", required=True
    )

    # Attribute sensitive
    attribute_sensitive = fields.Bool(
        example=True, description="Sensitivity of the attribute", required=False
    )

    # Attribute category
    attribute_category = fields.Str(
        example="Personal", description="Category of the attribute", required=False
    )

    # Attribute description
    attribute_description = fields.Str(
        required=True,
        description="The description of the attribute.",
        example="Name of the customer",
    )

    restrictions = fields.List(
        fields.Nested(DataAgreementPersonalDataRestrictionSchema), required=False
    )


class DataAgreementEvent(BaseModel):
    """
    Data agreement event model class
    """

    class Meta:
        # Schema class
        schema_class = "DataAgreementEventSchema"

        # Unknown fields are excluded
        unknown = EXCLUDE

    def __init__(
        self,
        *,
        event_id: str = None,
        time_stamp: str = None,
        principle_did: str = None,
        state: str = None,
        **kwargs,
    ):
        # Call parent constructor
        super().__init__(**kwargs)

        # Set attributes
        self.event_id = event_id
        self.time_stamp = time_stamp
        self.principle_did = principle_did
        self.state = state


class DataAgreementEventSchema(BaseModelSchema):
    """
    Data agreement event schema class
    """

    class Meta:
        # Model class
        model_class = DataAgreementEvent

    event_id = fields.Str(
        data_key="id",
        example="did:mydata:z6MkfiSdYhnLnS6jfwSf2yS2CiwwjZGmFUFL5QbyL2Xu8z2E",
        description="Data agreement event identifier",
    )

    # Time stamp
    time_stamp = fields.Str(
        data_key="time_stamp",
        example=str(
            datetime.datetime.utcnow().replace(tzinfo=datetime.timezone.utc).isoformat()
        ),
        description="Data agreement event timestamp in ISO 8601 UTC date time format",
        dump_only=True,
    )

    # Principle DID
    principle_did = fields.Str(
        data_key="data_subject_did",
        description="MyData decentralised identifier",
        dump_only=True,
        **MYDATA_DID,
    )

    # State
    state = fields.Str(
        description="State of the event", example="capture", dump_only=True
    )


class DataAgreementV1(BaseModel):
    """
    Data agreement model class
    """

    class Meta:
        # Schema class
        schema_class = "DataAgreementV1Schema"

    def __init__(
        self,
        *,
        context: str = None,
        data_agreement_template_id: str = None,
        data_agreement_template_version: int = None,
        pii_controller_name: str = None,
        pii_controller_url: str = None,
        usage_purpose: str = None,
        usage_purpose_description: str = None,
        legal_basis: str = None,
        method_of_use: str = None,
        data_policy: DataAgreementDataPolicy = None,
        personal_data: List[DataAgreementPersonalData] = None,
        dpia: DataAgreementDPIA = None,
        **kwargs,
    ):
        """
        Initialize data agreement model
        """

        # Call parent constructor
        super().__init__(**kwargs)

        # Set attributes
        self.context = context
        self.data_agreement_template_id = data_agreement_template_id
        self.data_agreement_template_version = data_agreement_template_version
        self.pii_controller_name = pii_controller_name
        self.pii_controller_url = pii_controller_url
        self.usage_purpose = usage_purpose
        self.usage_purpose_description = usage_purpose_description
        self.legal_basis = legal_basis
        self.method_of_use = method_of_use
        self.data_policy = data_policy
        self.personal_data = personal_data
        self.dpia = dpia


class DataAgreementV1Schema(BaseModelSchema):
    """
    Data agreement schema class
    """

    class Meta:
        # Model class
        model_class = DataAgreementV1

        # Unknown fields are excluded
        unknown = EXCLUDE

    # Data agreement schema context i.e. which schema is used
    context = fields.Str(
        data_key="@context",
        example="https://raw.githubusercontent.com/decentralised-dataexchange/automated-data-agreements/main/interface-specs/data-agreement-schema/v1/data-agreement-schema-context.jsonld",
        description="Context of the schema",
        required=True,
    )

    # TODO: "data_agreement_id" and "data_agreement_version" are only
    # TODO: ... generated when the data agreement is presented to the end user
    # TODO: ... therefore they will only be part of DataAgreement "instance"
    # # Data agreement instance identifier
    # # i.e. identifier of the "capture" data agreement instance
    # data_agreement_id = fields.Str(
    #     data_key="id",
    #     example=UUIDFour.EXAMPLE,
    #     description="Data agreement identifier",
    #     required=False,
    # )

    # # Data agreement instance version
    # # i.e. version of the "capture" data agreement instance
    # data_agreement_version = fields.Str(
    #     data_key="version",
    #     example="v1.0",
    #     description="Data agreement version",
    #     dump_only=True,
    # )

    # Data agreement template identifier
    # i.e. identifier of the "prepared" data agreement template
    data_agreement_template_id = fields.Str(
        data_key="template_id",
        example=UUIDFour.EXAMPLE,
        description="Data agreement template identifier",
    )

    # Data agreement template version
    # i.e. version of the "prepared" data agreement template
    data_agreement_template_version = fields.Int(
        data_key="template_version",
        example=1,
        description="Data agreement template version",
    )

    @validates("pii_controller_name")
    def validate_pii_controller_name(self, value):
        """
        Validate data agreement schema pii controller name
        """
        if len(value) < 3:
            raise ValidationError(
                f"PII controller name must be at least 3 characters long."
            )
        if len(value) > 100:
            raise ValidationError(
                f"PII controller name must be at most 100 characters long."
            )

    # Data agreement data controller name
    # i.e. Organization name of the data controller
    pii_controller_name = fields.Str(
        data_key="data_controller_name",
        example="Happy Shopping AB",
        description="PII controller name",
        required=True,
    )

    @validates("pii_controller_url")
    def validate_pii_controller_url(self, value):
        """
        Validate data agreement schema pii controller url
        """
        if not validators.url(value):
            raise ValidationError(f"Provided PII controller URL is not valid.")

    # Data agreement data controller URL
    pii_controller_url = fields.Str(
        data_key="data_controller_url",
        example="https://www.happyshopping.com",
        description="PII controller URL",
    )

    @validates("usage_purpose")
    def validate_usage_purpose(self, value):
        """
        Validate data agreement schema usage purpose
        """
        if len(value) < 3:
            raise ValidationError(f"Usage purpose must be at least 3 characters long.")
        if len(value) > 100:
            raise ValidationError(f"Usage purpose must be at most 100 characters long.")

    # Data agreement usage purpose
    usage_purpose = fields.Str(
        data_key="purpose",
        example="Customized shopping experience",
        description="Usage purpose title",
        required=True,
    )

    @validates("usage_purpose_description")
    def validate_usage_purpose_description(self, value):
        """
        Validate data agreement schema usage purpose description
        """
        if len(value) < 3:
            raise ValidationError(
                f"Usage purpose description must be at least 3 characters long."
            )
        if len(value) > 500:
            raise ValidationError(
                f"Usage purpose description must be at most 500 characters long."
            )

    # Data agreement usage purpose description
    usage_purpose_description = fields.Str(
        data_key="purpose_description",
        example="Collecting user data for offering custom tailored shopping experience",
        description="Usage purpose description",
        required=True,
    )

    # Data agreement legal basis
    legal_basis = fields.Str(
        data_key="lawful_basis",
        example="consent",
        description="Legal basis of processing",
        required=True,
        validate=validate.OneOf(
            [
                "consent",
                "legal_obligation",
                "contract",
                "vital_interest",
                "public_task",
                "legitimate_interest",
            ]
        ),
    )

    # Data agreement method of use (i.e. how the data is used)
    # 2 method of use: "data-source" and "data-using-service"
    method_of_use = fields.Str(
        data_key="method_of_use",
        example="data-using-service",
        description="Method of use (or data exchange mode)",
        required=True,
        validate=validate.OneOf(
            [
                "data-source",
                "data-using-service",
            ]
        ),
    )

    # Data agreement data policy
    data_policy = fields.Nested(DataAgreementDataPolicySchema, required=True)

    # Data agreement personal data (attributes)
    personal_data = fields.List(
        fields.Nested(DataAgreementPersonalDataSchema),
        required=True,
        validate=validate.Length(min=1),
    )

    # Data agreement DPIA metadata
    dpia = fields.Nested(DataAgreementDPIASchema, required=False)

    # TODO: "event" list will be populated during the capture phase of DA lifecycle
    # event = fields.List(fields.Nested(
    #     DataAgreementEventSchema), dump_only=True)



from marshmallow import EXCLUDE, fields
from aries_cloudagent.messaging.agent_message import AgentMessage, AgentMessageSchema
from dexa_sdk.agreements.da.v1_0.models.da_instance_models import (
    DataAgreementInstanceSchema,
    DataAgreementInstanceModel
)
from ..message_types import DATA_AGREEMENT_NEGOTIATION_OFFER


class DataAgreementNegotiationOfferMessage(AgentMessage):
    """
    Message class for data agreement negotiation offer
    """

    class Meta:

        # Message type
        message_type = DATA_AGREEMENT_NEGOTIATION_OFFER

        # Message schema class
        schema_class = "DataAgreementNegotiationOfferMessageSchema"

    def __init__(
        self,
        *,
        body: DataAgreementInstanceModel,
        **kwargs
    ):
        """
        Initialize a DataAgreementNegotiationOfferMessage message instance.
        """
        super().__init__(**kwargs)

        # Message body
        self.body = body


class DataAgreementNegotiationOfferMessageSchema(AgentMessageSchema):
    """
    Schema class for data agreement negotiation offer message
    """

    class Meta:
        # The message class that this schema is for
        model_class = DataAgreementNegotiationOfferMessage

        # Unknown fields to exclude from the schema (handled by marshmallow)
        unknown = EXCLUDE

    # Message body
    body = fields.Nested(
        DataAgreementInstanceSchema,
        required=True
    )



from aries_cloudagent.messaging.base_handler import BaseHandler, BaseResponder, RequestContext

from ..messages.problem_report import DataAgreementProblemReport
from ..manager import ADAManager

import json


class DataAgreementProblemReportHandler(BaseHandler):
    """Handle for data-agreements/1.0/problem-report message"""

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """
        Message handler logic for data-agreements/1.0/problem-report message.
        """

        # Assert if received message is of type DataAgreementProblemReport
        assert isinstance(context.message, DataAgreementProblemReport)

        self._logger.info(
            "Received create-did message: \n%s",
            json.dumps(context.message.serialize(), indent=4)
        )

        # Initialize ADA manager
        ada_manager = ADAManager(context)

        await ada_manager.process_data_agreement_problem_report_message(
            data_agreement_problem_report_message=context.message,
            receipt=context.message_receipt,
        )



from marshmallow import fields
from aries_cloudagent.messaging.models.base import BaseModel, BaseModelSchema
from aries_cloudagent.messaging.valid import UUIDFour


class DataAgreementQrCodeInitiateBody(BaseModel):
    """Data Agreement Qr code initiate body"""

    class Meta:
        """Data Agreement Qr code initiate Body metadata"""

        schema_class = "DataAgreementQrCodeInitiateBodySchema"

    def __init__(
        self,
        *,
        qr_id: str = None,
        **kwargs
    ):
        """Data Agreement Qr Code Initiate Body init"""

        super().__init__(**kwargs)

        # Set attributes
        self.qr_id = qr_id


class DataAgreementQrCodeInitiateBodySchema(BaseModelSchema):
    """Data Agreement Qr Code Initiate Body schema"""

    class Meta:
        """Data Agreement Qr Code Initiate Body schema metadata"""

        model_class = DataAgreementQrCodeInitiateBody

    # Data agreement id
    qr_id = fields.Str(
        example=UUIDFour.EXAMPLE,
        description="Qr code identifier"
    )



from aries_cloudagent.messaging.decorators.base import BaseDecoratorSet
from aries_cloudagent.messaging.decorators.localization_decorator import LocalizationDecorator
from aries_cloudagent.messaging.decorators.signature_decorator import SignatureDecorator
from aries_cloudagent.messaging.decorators.thread_decorator import ThreadDecorator
from aries_cloudagent.messaging.decorators.trace_decorator import TraceDecorator
from aries_cloudagent.messaging.decorators.timing_decorator import TimingDecorator
from aries_cloudagent.messaging.decorators.transport_decorator import TransportDecorator

from .data_agreement_context_decorator import DataAgreementContextDecorator

PATCHED_DECORATOR_MODELS = {
    "l10n": LocalizationDecorator,
    "sig": SignatureDecorator,
    "thread": ThreadDecorator,
    "trace": TraceDecorator,
    "timing": TimingDecorator,
    "transport": TransportDecorator,
    "data-agreement-context": DataAgreementContextDecorator,
}


class PatchedDecoratorSet(BaseDecoratorSet):
    """Patched decorator set implementation."""

    def __init__(self, models: dict = None):
        """Initialize the decorator set."""
        super().__init__(PATCHED_DECORATOR_MODELS if models is None else models)



# flake8: noqa
from aries_cloudagent.protocols.didcomm_prefix import DIDCommPrefix

# DIDComm protocol specification for Automated Data Agreements
SPEC_URI = (
    "https://github.com/decentralised-dataexchange/automated-data-agreements"
)

# Message types

# Message types for MyDATA DID CRUD operations (ADA RFC 0001 - MyData DID Protocol 1.0)
READ_DID = f"mydata-did/1.0/read-did"
READ_DID_RESPONSE = f"mydata_did/1.0/read-did-response"
MYDATA_DID_PROBLEM_REPORT = f"mydata-did/1.0/problem-report"

# Message types for Data Agreement CRUD operations (ADA RFC 0002 - Data Agreement Protocol 1.0)
READ_DATA_AGREEMENT = f"data-agreements/1.0/read-data-agreement"
READ_DATA_AGREEMENT_RESPONSE = f"data-agreements/1.0/read-data-agreement-response"
DATA_AGREEMENT_PROBLEM_REPORT = f"data-agreements/1.0/problem-report"

# Message types for ADA RFC 0003 - Data Agreement Negotiation Protocol 1.0
DATA_AGREEMENT_NEGOTIATION_OFFER = f"data-agreement-negotiation/1.0/offer"
DATA_AGREEMENT_NEGOTIATION_REJECT = f"data-agreement-negotiation/1.0/reject"
DATA_AGREEMENT_NEGOTIATION_ACCEPT = f"data-agreement-negotiation/1.0/accept"
DATA_AGREEMENT_NEGOTIATION_RECEIPT = f"data-agreement-negotiation/1.0/receipt"
DATA_AGREEMENT_NEGOTIATION_PROBLEM_REPORT = f"data-agreement-negotiation/1.0/problem-report"

# Message types for ADA RFC 0005 - Data Agreement Termination Protocol 1.0
DATA_AGREEMENT_TERMINATION_TERMINATE = f"data-agreement-termination/1.0/terminate"
DATA_AGREEMENT_TERMINATION_TERMINATE_ACK = f"data-agreement-termination/1.0/terminate-ack"
DATA_AGREEMENT_TERMINATION_PROBLEM_REPORT = f"data-agreement-termination/1.0/problem-report"

# Message types for iGrant.io specific messages.

# Data agreement QR code workflow initiate message.
DATA_AGREEMENT_QR_CODE_WORKFLOW_INITIATE = f"data-agreement-qr-code/1.0/initiate"
DATA_AGREEMENT_QR_CODE_WORKFLOW_PROBLEM_REPORT = f"data-agreement-qr-code/1.0/problem-report"

# JSON-LD functions protocol.
JSON_LD_PROCESSED_DATA = f"json-ld/1.0/processed-data"
JSON_LD_PROCESSED_RESPONSE_DATA = f"json-ld/1.0/processed-data-response"
JSON_LD_PROBLEM_REPORT = f"json-ld/1.0/problem-report"

# Data controller protocol
DATA_CONTROLLER_DETAILS = f"data-controller/1.0/details"
DATA_CONTROLLER_DETAILS_RESPONSE = f"data-controller/1.0/details-response"

# Existing connections protocol.
EXISTING_CONNECTIONS = f"connections/1.0/exists"

# Protocol package path
PROTOCOL_PACKAGE = "mydata_did.v1_0"

# Message type mappings to their corresponding class
MESSAGE_TYPES = DIDCommPrefix.qualify_all(
    {
        READ_DID: (
            f"{PROTOCOL_PACKAGE}.messages.read_did.ReadDIDMessage"
        ),
        READ_DID_RESPONSE: (
            f"{PROTOCOL_PACKAGE}.messages.read_did_response.ReadDIDResponseMessage"
        ),
        MYDATA_DID_PROBLEM_REPORT: (
            f"{PROTOCOL_PACKAGE}.messages.problem_report.MyDataDIDProblemReportMessage"
        ),
        READ_DATA_AGREEMENT: (
            f"{PROTOCOL_PACKAGE}.messages.read_data_agreement.ReadDataAgreement"
        ),
        READ_DATA_AGREEMENT_RESPONSE: (
            f"{PROTOCOL_PACKAGE}.messages.read_data_agreement_response.ReadDataAgreementResponse"
        ),
        DATA_AGREEMENT_PROBLEM_REPORT: (
            f"{PROTOCOL_PACKAGE}.messages.problem_report.DataAgreementProblemReport"
        ),
        DATA_AGREEMENT_NEGOTIATION_RECEIPT: (
            f"{PROTOCOL_PACKAGE}.messages.da_negotiation_receipt.DataAgreementNegotiationReceiptMessage"
        ),
        DATA_AGREEMENT_NEGOTIATION_REJECT: (
            f"{PROTOCOL_PACKAGE}.messages.data_agreement_reject.DataAgreementNegotiationRejectMessage"
        ),
        DATA_AGREEMENT_NEGOTIATION_PROBLEM_REPORT: (
            f"{PROTOCOL_PACKAGE}.messages.problem_report.DataAgreementNegotiationProblemReport"
        ),
        DATA_AGREEMENT_TERMINATION_TERMINATE: (
            f"{PROTOCOL_PACKAGE}.messages.data_agreement_terminate.DataAgreementTerminationTerminateMessage"
        ),
        DATA_AGREEMENT_TERMINATION_TERMINATE_ACK: (
            f"{PROTOCOL_PACKAGE}.messages.data_agreement_terminate_ack.DataAgreementTerminationAck"
        ),
        DATA_AGREEMENT_TERMINATION_PROBLEM_REPORT: (
            f"{PROTOCOL_PACKAGE}.messages.problem_report.DataAgreementTerminationProblemReport"
        ),
        DATA_AGREEMENT_QR_CODE_WORKFLOW_INITIATE: (
            f"{PROTOCOL_PACKAGE}.messages.data_agreement_qr_code_initiate.DataAgreementQrCodeInitiateMessage"
        ),
        DATA_AGREEMENT_QR_CODE_WORKFLOW_PROBLEM_REPORT: (
            f"{PROTOCOL_PACKAGE}.messages.data_agreement_qr_code_problem_report.DataAgreementQrCodeProblemReport"
        ),
        JSON_LD_PROCESSED_DATA: (
            f"{PROTOCOL_PACKAGE}.messages.json_ld_processed.JSONLDProcessedMessage"
        ),
        JSON_LD_PROCESSED_RESPONSE_DATA: (
            f"{PROTOCOL_PACKAGE}.messages.json_ld_processed_response.JSONLDProcessedResponseMessage"
        ),
        JSON_LD_PROBLEM_REPORT: (
            f"{PROTOCOL_PACKAGE}.messages.json_ld_problem_report.JSONLDProblemReport"
        ),
        DATA_CONTROLLER_DETAILS: (
            f"{PROTOCOL_PACKAGE}.messages.data_controller_details.DataControllerDetailsMessage"
        ),
        DATA_CONTROLLER_DETAILS_RESPONSE: (
            f"{PROTOCOL_PACKAGE}.messages.data_controller_details_response.DataControllerDetailsResponseMessage"
        ),
        EXISTING_CONNECTIONS: (
            f"{PROTOCOL_PACKAGE}.messages.existing_connections.ExistingConnectionsMessage"
        ),
    },
)



from aries_cloudagent.messaging.base_handler import BaseHandler, BaseResponder, RequestContext

from ..messages.read_did import ReadDIDMessage
from ..manager import ADAManager

import json


class ReadDIDHandler(BaseHandler):
    """
    Read DID message handler
    """

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """
        Message handler logic for read-did request.
        """

        # Assert if received message is of type ReadDIDMessage
        assert isinstance(context.message, ReadDIDMessage)

        self._logger.info(
            "Received read-did message: \n%s",
            json.dumps(context.message.serialize(), indent=4)
        )

        # Initialize ADA manager
        mgr = ADAManager(context)

        # Process read-did message
        await mgr.process_read_did_message(context.message, context.message_receipt)



from aries_cloudagent.messaging.agent_message import AgentMessage, AgentMessageSchema
from marshmallow import EXCLUDE, fields

from ..message_types import DATA_CONTROLLER_DETAILS_RESPONSE, PROTOCOL_PACKAGE
from ..models.data_controller_model import DataController, DataControllerSchema

# Handler class for data controller details response
HANDLER_CLASS = (
    f"{PROTOCOL_PACKAGE}.handlers"
    ".data_controller_details_response_handler.DataControllerDetailsResponseHandler"
)


class DataControllerDetailsResponseMessage(AgentMessage):
    """
    Message class for data controller details response message.
    """

    class Meta:
        # Handler class that can handle this message
        handler_class = HANDLER_CLASS

        # Message type
        message_type = DATA_CONTROLLER_DETAILS_RESPONSE

        # Message schema class
        schema_class = "DataControllerDetailsResponseMessageSchema"

    def __init__(
        self,
        *,
        body: DataController,
        **kwargs
    ):
        """
        Initialize a DataControllerDetailsResponseMessage message instance.
        """
        super().__init__(**kwargs)

        # Data controller details
        self.body = body


class DataControllerDetailsResponseMessageSchema(AgentMessageSchema):
    """
    Schema class for data controller details response message
    """

    class Meta:
        # The message class that this schema is for
        model_class = DataControllerDetailsResponseMessage

        # Unknown fields to exclude from the schema (handled by marshmallow)
        unknown = EXCLUDE

    # body
    body = fields.Nested(DataControllerSchema)



"""Sign and verify functions for json-ld based data agreements."""

import json
import typing

from aries_cloudagent.wallet.util import (
    b58_to_bytes,
    b64_to_bytes,
    b64_to_str,
    bytes_to_b58,
    bytes_to_b64,
    str_to_b64,
)

from aries_cloudagent.wallet.base import BaseWallet

from .create_verify_data import create_verify_data


MULTIBASE_B58_BTC = "z"
MULTICODEC_ED25519_PUB = b"\xed"


def did_mydata(verkey: str) -> str:
    """Qualify verkey into DID MyData if need be."""

    if verkey.startswith(f"did:mydata:{MULTIBASE_B58_BTC}"):
        return verkey

    return f"did:mydata:{MULTIBASE_B58_BTC}" + bytes_to_b58(
        MULTICODEC_ED25519_PUB + b58_to_bytes(verkey)
    )


def b64encode(str):
    """Url Safe B64 Encode."""
    return str_to_b64(str, urlsafe=True, pad=False)


def b64decode(bytes):
    """Url Safe B64 Decode."""
    return b64_to_str(bytes, urlsafe=True)


def create_jws(encoded_header, verify_data):
    """Compose JWS."""
    return (encoded_header + ".").encode("utf-8") + verify_data


async def jws_sign(verify_data, verkey, wallet):
    """Sign JWS."""

    header = {"alg": "EdDSA", "b64": False, "crit": ["b64"]}

    encoded_header = b64encode(json.dumps(header))

    jws_to_sign = create_jws(encoded_header, verify_data)

    signature = await wallet.sign_message(jws_to_sign, verkey)

    encoded_signature = bytes_to_b64(signature, urlsafe=True, pad=False)

    return encoded_header + ".." + encoded_signature


def verify_jws_header(header):
    """Check header requirements."""

    if (
        not (
            header["alg"] == "EdDSA"
            and header["b64"] is False
            and isinstance(header["crit"], list)
            and len(header["crit"]) == 1
            and header["crit"][0] == "b64"
        )
        and len(header) == 3
    ):
        raise Exception("Invalid JWS header parameters for Ed25519Signature2018.")


async def jws_verify(verify_data, signature, public_key, wallet):
    """Detatched jws verify handling."""

    encoded_header, _, encoded_signature = signature.partition("..")
    decoded_header = json.loads(b64decode(encoded_header))

    verify_jws_header(decoded_header)

    decoded_signature = b64_to_bytes(encoded_signature, urlsafe=True)

    jws_to_verify = create_jws(encoded_header, verify_data)

    verified = await wallet.verify_message(jws_to_verify, decoded_signature, public_key)

    return verified


async def sign_data_agreement(data_agreement, signature_options, verkey, wallet):
    """Sign data agreement."""

    proof_chain = False
    if (
        "proof" in data_agreement
    ):
        # Detected the document is being signed more than once, 
        # therefore expected output is a proof chain of 2 elements.

        proof_chain = True

    framed, verify_data_hex_string = create_verify_data(data_agreement, signature_options, proof_chain)

    verify_data_bytes = bytes.fromhex(verify_data_hex_string)

    jws = await jws_sign(verify_data_bytes, verkey, wallet)

    if "proofChain" not in data_agreement:
        if not proof_chain:
            # For single proof

            document_with_proof = {**data_agreement, "proof": {**signature_options, "proofValue": jws}}
        else:
            # For proof chain with 2 elements

            old_proof = data_agreement.pop("proof", None)
            new_proof = {**signature_options, "proofValue": jws}
            document_with_proof = {**data_agreement, "proofChain": [old_proof, new_proof]}
    else:

        # For proof chain with more than 2 elements

        document_with_proof = {**data_agreement}
        document_with_proof["proofChain"].append({
            **signature_options, "proofValue": jws
        })

    return document_with_proof


async def verify_data_agreement(doc, verkey, wallet, drop_proof_chain:bool = True):
    """Verify data agreement."""

    proof_chain = False
    old_proof = None
    new_proof = None
    if (
        "proofChain" in doc
    ):
        # Detected the document is being signed more than once, therefore it is a proof chain.
        proof_chain = True

        if drop_proof_chain:
            # For proof chain with 2 elements

            old_proof = doc["proofChain"][0]
            new_proof = doc["proofChain"][1]

            doc.pop("proofChain", None)
            
            doc["proof"] = old_proof
        else:
            # For proof chain with more than 2 elements

            new_proof = doc["proofChain"][-1]

            doc["proofChain"] = doc["proofChain"][:-1]

    framed, verify_data_hex_string = create_verify_data(doc, doc["proof"] if not proof_chain else new_proof, proof_chain)

    verify_data_bytes = bytes.fromhex(verify_data_hex_string)

    valid = await jws_verify(verify_data_bytes, framed["proof"]["proofValue"] if not proof_chain else new_proof["proofValue"], verkey, wallet)

    return valid

async def verify_data_agreement_with_proof_chain(doc: dict = None, verkeys: typing.List[str] = None, wallet: BaseWallet = None) -> bool:

    proof_chain = doc.pop("proofChain", None)
    genesis_proof = proof_chain[0]

    genesis_doc = {**doc, "proof": genesis_proof}
    genesis_doc["event"] = genesis_doc["event"][0]

    current_doc = {**doc, "proofChain": proof_chain}

    genesis_valid = await verify_data_agreement(genesis_doc.copy(), verkeys[0], wallet)

    current_valid = await verify_data_agreement(current_doc.copy(), verkeys[1], wallet)

    return genesis_valid and current_valid



from marshmallow import fields
from aries_cloudagent.messaging.models.base import BaseModel, BaseModelSchema


class ExistingConnectionsBody(BaseModel):
    """Existing connections message body"""

    class Meta:
        """Existing connection message body metadata"""

        schema_class = "ExistingConnectionsBodySchema"

    def __init__(
        self,
        *,
        theirdid: str = None,
        **kwargs
    ):
        """ExistingConnectionsBody init"""

        super().__init__(**kwargs)

        # Set attributes
        self.theirdid = theirdid


class ExistingConnectionsBodySchema(BaseModelSchema):
    """Existing connections message body schema"""

    class Meta:
        """ExistingConnectionsBodySchema metadata"""

        model_class = ExistingConnectionsBody

    # Pairwise DID of the existing connection.
    theirdid = fields.Str()



import typing
import datetime

from marshmallow import EXCLUDE, fields, validate, validates
from marshmallow.exceptions import ValidationError
from aries_cloudagent.messaging.models.base import BaseModel, BaseModelSchema
from aries_cloudagent.messaging.valid import UUIDFour

from ..models.data_agreement_model import (
    DataAgreementDataPolicy,
    DataAgreementDataPolicySchema,
    DataAgreementPersonalData,
    DataAgreementPersonalDataSchema,
    DataAgreementDPIA,
    DataAgreementDPIASchema,
    DATA_AGREEMENT_V1_SCHEMA_CONTEXT,
)

from ..models.data_agreement_negotiation_offer_model import (
    DataAgreementEvent,
    DataAgreementProof,
    DataAgreementProofSchema,
    DataAgreementEventSchema
)

from ..utils.jsonld import ED25519_2018_CONTEXT_URL


class DataAgreementInstance(BaseModel):
    """Data Agreement instance Body"""

    class Meta:
        """Data Agreement instance metadata"""

        schema_class = "DataAgreementInstanceSchema"

    def __init__(
        self,
        *,
        context: typing.List[str],
        data_agreement_id: str = None,
        data_agreement_version: int = None,
        data_agreement_template_id: str = None,
        data_agreement_template_version: int = None,
        pii_controller_name: str = None,
        pii_controller_url: str = None,
        usage_purpose: str = None,
        usage_purpose_description: str = None,
        legal_basis: str = None,
        method_of_use: str = None,
        data_policy: DataAgreementDataPolicy = None,
        personal_data: typing.List[DataAgreementPersonalData] = None,
        dpia: DataAgreementDPIA = None,
        event: typing.List[DataAgreementEvent] = None,
        proof_chain: typing.List[DataAgreementProof] = None,
        principle_did: str = None,
        proof:  DataAgreementProof = None,
        **kwargs
    ):
        """Data Agreement instance init"""

        super().__init__(**kwargs)

        # Set attributes
        self.context = context
        self.data_agreement_id = data_agreement_id
        self.data_agreement_version = data_agreement_version
        self.data_agreement_template_id = data_agreement_template_id
        self.data_agreement_template_version = data_agreement_template_version
        self.pii_controller_name = pii_controller_name
        self.pii_controller_url = pii_controller_url
        self.usage_purpose = usage_purpose
        self.usage_purpose_description = usage_purpose_description
        self.legal_basis = legal_basis
        self.method_of_use = method_of_use
        self.data_policy = data_policy
        self.personal_data = personal_data
        self.dpia = dpia
        self.proof_chain = proof_chain
        self.principle_did = principle_did
        self.event = event
        self.proof = proof


class DataAgreementInstanceSchema(BaseModelSchema):
    """Data Agreement instance schema"""

    class Meta:
        """Data Agreement instance metadata"""

        model_class = DataAgreementInstance

    # Context
    context = fields.List(
        fields.Str(),
        example=[DATA_AGREEMENT_V1_SCHEMA_CONTEXT, ED25519_2018_CONTEXT_URL],
        data_key="@context",
        description="Context",
    )

    # Data agreement id
    data_agreement_id = fields.Str(
        data_key="id",
        example=UUIDFour.EXAMPLE,
        description="Data agreement identifier"
    )

    # Data agreement version
    data_agreement_version = fields.Int(
        data_key="version",
        example=1,
        description="Data agreement version"
    )

    # Data agreement template id
    data_agreement_template_id = fields.Str(
        data_key="template_id",
        example=UUIDFour.EXAMPLE,
        description="Data agreement template identifier"
    )

    # Data agreement template version
    data_agreement_template_version = fields.Int(
        data_key="template_version",
        example=1,
        description="Data agreement template version"
    )

    # Data agreement data controller name
    # i.e. Organization name of the data controller
    pii_controller_name = fields.Str(
        data_key="data_controller_name",
        example="Happy Shopping AB",
        description="PII controller name"
    )

    # Data agreement data controller URL
    pii_controller_url = fields.Str(
        data_key="data_controller_url",
        example="https://www.happyshopping.com",
        description="PII controller URL"
    )

    # Data agreement usage purpose
    usage_purpose = fields.Str(
        data_key="purpose",
        example="Customized shopping experience",
        description="Usage purpose title",
        required=True
    )

    # Data agreement usage purpose description
    usage_purpose_description = fields.Str(
        data_key="purpose_description",
        example="Collecting user data for offering custom tailored shopping experience",
        description="Usage purpose description",
        required=True
    )

    # Data agreement legal basis
    legal_basis = fields.Str(
        data_key="lawful_basis",
        example="consent",
        description="Legal basis of processing",
        required=True,
        validate=validate.OneOf(
            [
                "consent",
                "legal_obligation",
                "contract",
                "vital_interest",
                "public_task",
                "legitimate_interest",

            ]
        )
    )

    # Data agreement method of use (i.e. how the data is used)
    # 2 method of use: "data-source" and "data-using-service"
    method_of_use = fields.Str(
        data_key="method_of_use",
        example="data-using-service",
        description="Method of use (or data exchange mode)",
        required=True,
        validate=validate.OneOf(
            [
                "data-source",
                "data-using-service",

            ]
        )
    )

    # Data agreement data policy
    data_policy = fields.Nested(
        DataAgreementDataPolicySchema,
        required=True,
        description="Data policy"
    )

    # Data agreement personal data (attributes)
    personal_data = fields.List(
        fields.Nested(DataAgreementPersonalDataSchema),
        required=True,
        description="Personal data (attributes)"
    )

    # Data agreement DPIA metadata
    dpia = fields.Nested(
        DataAgreementDPIASchema,
        description="DPIA metadata"
    )

    # Data agreement events
    event = fields.List(fields.Nested(DataAgreementEventSchema))

    # Data agreement proof chain
    proof_chain = fields.List(
        fields.Nested(DataAgreementProofSchema),
        description="Data agreement proof chain",
        data_key="proofChain"
    )

    # Data agreement principle did
    principle_did = fields.Str(
        data_key="data_subject_did",
        example="did:mydata:123456789abcdefghi",
        description="Principle did"
    )

    # Data agreement proof
    proof = fields.Nested(DataAgreementProofSchema, required=False)



from marshmallow import fields
from aries_cloudagent.messaging.models.base import BaseModel, BaseModelSchema
from aries_cloudagent.messaging.valid import UUIDFour


class JSONLDProcessedResponseBody(BaseModel):
    """json-ld/1.0/processed-data-response message body"""

    class Meta:
        """json-ld/1.0/processed-data-response message body metadata"""

        schema_class = "JSONLDProcessedResponseBodySchema"

    def __init__(
        self,
        *,
        framed_base64: str = None,
        combined_hash_base64: str = None,
        **kwargs
    ):
        """json-ld/1.0/processed-data-response message body init"""

        super().__init__(**kwargs)

        # Set attributes
        self.framed_base64 = framed_base64
        self.combined_hash_base64 = combined_hash_base64


class JSONLDProcessedResponseBodySchema(BaseModelSchema):
    """json-ld/1.0/processed-data-response message body schema"""

    class Meta:
        """json-ld/1.0/processed-data-response message body schema metadata"""

        model_class = JSONLDProcessedResponseBody

    # JSON-LD framed data
    framed_base64 = fields.Str(
        description="JSON-LD framed data",
    )

    # Combined hash ( JSON-LD data + signature options )
    combined_hash_base64 = fields.Str(
        description="Combined hash ( JSON-LD data + signature options )",
    )



from aries_cloudagent.messaging.agent_message import AgentMessage, AgentMessageSchema
from marshmallow import EXCLUDE, fields

from ..message_types import DATA_AGREEMENT_QR_CODE_WORKFLOW_INITIATE, PROTOCOL_PACKAGE
from ..models.data_agreement_qr_code_initiate_model import (
    DataAgreementQrCodeInitiateBody,
    DataAgreementQrCodeInitiateBodySchema
)

# Handler class for data-agreement-qr-code/1.0/initiate message
HANDLER_CLASS = (
    f"{PROTOCOL_PACKAGE}.handlers"
    ".data_agreement_qr_code_initiate_handler.DataAgreementQrCodeInitiateHandler"
)


class DataAgreementQrCodeInitiateMessage(AgentMessage):
    """
    Message class for data agreement Qr code workflow initiate
    """

    class Meta:

        # Handler class that can handle this message
        handler_class = HANDLER_CLASS

        # Message type
        message_type = DATA_AGREEMENT_QR_CODE_WORKFLOW_INITIATE

        # Message schema class
        schema_class = "DataAgreementQrCodeInitiateMessageSchema"

    def __init__(
        self,
        *,
        body: DataAgreementQrCodeInitiateBody,
        **kwargs
    ):
        """
        Initialize a DataAgreementQrCodeInitiateMessage message instance.
        """
        super().__init__(**kwargs)

        # Message body
        self.body = body


class DataAgreementQrCodeInitiateMessageSchema(AgentMessageSchema):
    """
    Schema class for data agreement qr code initiate message
    """

    class Meta:
        # The message class that this schema is for
        model_class = DataAgreementQrCodeInitiateMessage

        # Unknown fields to exclude from the schema (handled by marshmallow)
        unknown = EXCLUDE

    # Message body
    body = fields.Nested(
        DataAgreementQrCodeInitiateBodySchema,
        required=True
    )



from aries_cloudagent.messaging.base_handler import BaseHandler, BaseResponder, RequestContext

from ..messages.json_ld_processed_response import JSONLDProcessedResponseMessage
from ..manager import ADAManager

import json


class JSONLDProcessedResponseHandler(BaseHandler):
    """Handle for json-ld/1.0/processed message"""

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """
        Message handler logic for json-ld/1.0/processed-data-response message.
        """

        # Assert if received message is of type JSONLDProcessedResponseMessage
        assert isinstance(context.message, JSONLDProcessedResponseMessage)

        self._logger.info(
            "Received json-ld/1.0/processed-data-response message: \n%s",
            json.dumps(context.message.serialize(), indent=4)
        )




from aries_cloudagent.messaging.agent_message import AgentMessage, AgentMessageSchema
from marshmallow import EXCLUDE, fields

from ..message_types import DATA_AGREEMENT_TERMINATION_TERMINATE, PROTOCOL_PACKAGE
from ..utils.regex import MYDATA_DID
from ..models.data_agreement_termination_terminate_model import (
    DataAgreementTerminationTerminateBody,
    DataAgreementTerminationTerminateBodySchema
)

# Handler class for data agreement terminate message
HANDLER_CLASS = (
    f"{PROTOCOL_PACKAGE}.handlers"
    ".data_agreement_termination_terminate_handler.DataAgreementTerminationTerminateMessageHandler"
)

class DataAgreementTerminationTerminateMessage(AgentMessage):
    """
    Message class for data agreement termination terminate message.
    """

    class Meta:
        # Handler class that can handle this message
        handler_class = HANDLER_CLASS

        # Message type
        message_type = DATA_AGREEMENT_TERMINATION_TERMINATE

        # Message schema class
        schema_class = "DataAgreementTerminationTerminateMessageSchema"

    def __init__(
        self,
        *,
        from_did,
        to_did,
        created_time,
        body: DataAgreementTerminationTerminateBody,
        **kwargs
    ):
        """
        Initialize a DataAgreementTerminationTerminateMessage message instance.
        """
        super().__init__(**kwargs)

        # Sender DID
        self.from_did = from_did

        # Recipient DID
        self.to_did = to_did

        # The time the message was created
        self.created_time = created_time

        # Message body
        self.body = body


class DataAgreementTerminationTerminateMessageSchema(AgentMessageSchema):
    """
    Schema class for data agreement termination terminate message
    """

    class Meta:
        # The message class that this schema is for
        model_class = DataAgreementTerminationTerminateMessage

        # Unknown fields to exclude from the schema (handled by marshmallow)
        unknown = EXCLUDE

    # From DID
    from_did = fields.Str(data_key="from", **MYDATA_DID)

    # To DID
    to_did = fields.Str(data_key="to", **MYDATA_DID)

    # Created time
    created_time = fields.Str(data_key="created_time")

    # Message body
    body = fields.Nested(
        DataAgreementTerminationTerminateBodySchema, 
        required=True
    )



"""JSON Web Encryption utilities."""

import json

from collections import OrderedDict
from typing import Any, Iterable, List, Mapping, Optional, Union

from marshmallow import fields, Schema, ValidationError

from .wallet.util import b64_to_bytes, bytes_to_b64

IDENT_ENC_KEY = "encrypted_key"
IDENT_HEADER = "header"
IDENT_PROTECTED = "protected"
IDENT_RECIPIENTS = "recipients"


def b64url(value: Union[bytes, str]) -> str:
    """Encode a string or bytes value as unpadded base64-URL."""
    if isinstance(value, str):
        value = value.encode("utf-8")
    return bytes_to_b64(value, urlsafe=True, pad=False)


def from_b64url(value: str) -> bytes:
    """Decode an unpadded base64-URL value."""
    return b64_to_bytes(value, urlsafe=True)


class B64Value(fields.Str):
    """A marshmallow-compatible wrapper for base64-URL values."""

    def _serialize(self, value, attr, obj, **kwargs) -> Optional[str]:
        if value is None:
            return None
        if not isinstance(value, bytes):
            return TypeError("Expected bytes")
        return b64url(value)

    def _deserialize(self, value, attr, data, **kwargs) -> Any:
        value = super()._deserialize(value, attr, data, **kwargs)
        return from_b64url(value)


class JweSchema(Schema):
    """JWE envelope schema."""

    protected = fields.Str(required=True)
    unprotected = fields.Dict(required=False)
    ciphertext = B64Value(required=True)
    iv = B64Value(required=True)
    tag = B64Value(required=True)
    aad = B64Value(required=False)
    # flattened:
    header = fields.Dict(required=False)
    encrypted_key = B64Value(required=False)


class JweRecipientSchema(Schema):
    """JWE recipient schema."""

    encrypted_key = B64Value(required=True)
    header = fields.Dict(many=True, required=False)


class JweRecipient:
    """A single message recipient."""

    def __init__(self, *, encrypted_key: bytes, header: dict = None) -> "JweRecipient":
        """Initialize the JWE recipient."""
        self.encrypted_key = encrypted_key
        self.header = header or {}

    @classmethod
    def deserialize(cls, entry: Mapping[str, Any]) -> "JweRecipient":
        """Deserialize a JWE recipient from a mapping."""
        vals = JweRecipientSchema().load(entry)
        return cls(**vals)

    def serialize(self) -> dict:
        """Serialize the JWE recipient to a mapping."""
        ret = OrderedDict([("encrypted_key", b64url(self.encrypted_key))])
        if self.header:
            ret["header"] = self.header
        return ret


class JweEnvelope:
    """JWE envelope instance."""

    def __init__(
        self,
        *,
        protected: dict = None,
        protected_b64: bytes = None,
        unprotected: dict = None,
        ciphertext: bytes = None,
        iv: bytes = None,
        tag: bytes = None,
        aad: bytes = None,
    ):
        """Initialize a new JWE envelope instance."""
        self.protected = protected
        self.protected_b64 = protected_b64
        self.unprotected = unprotected or OrderedDict()
        self.ciphertext = ciphertext
        self.iv = iv
        self.tag = tag
        self.aad = aad
        self._recipients: List[JweRecipient] = []

    @classmethod
    def from_json(cls, message: Union[bytes, str]) -> "JweEnvelope":
        """Decode a JWE envelope from a JSON string or bytes value."""
        try:
            return cls._deserialize(JweSchema().loads(message))
        except json.JSONDecodeError:
            raise ValidationError("Invalid JWE: not JSON")

    @classmethod
    def deserialize(cls, message: Mapping[str, Any]) -> "JweEnvelope":
        """Deserialize a JWE envelope from a mapping."""
        return cls._deserialize(JweSchema().load(message))

    @classmethod
    def _deserialize(cls, parsed: Mapping[str, Any]) -> "JweEnvelope":
        protected_b64 = parsed[IDENT_PROTECTED]
        try:
            protected: dict = json.loads(from_b64url(protected_b64))
        except json.JSONDecodeError:
            raise ValidationError(
                "Invalid JWE: invalid JSON for protected headers"
            ) from None
        unprotected = parsed.get("unprotected") or dict()
        if protected.keys() & unprotected.keys():
            raise ValidationError("Invalid JWE: duplicate header")

        if IDENT_RECIPIENTS in protected:
            recips = [
                JweRecipient.deserialize(recip)
                for recip in protected.pop(IDENT_RECIPIENTS)
            ]
            if IDENT_ENC_KEY in protected or IDENT_HEADER in protected:
                raise ValidationError("Invalid JWE: flattened form with recipients")
        else:
            if IDENT_ENC_KEY not in protected:
                raise ValidationError("Invalid JWE: no recipients")
            header = protected.pop(IDENT_HEADER) if IDENT_HEADER in protected else None
            recips = [
                JweRecipient(
                    encrypted_key=from_b64url(protected.pop(IDENT_ENC_KEY)),
                    header=header,
                )
            ]

        inst = cls(
            protected=protected,
            protected_b64=protected_b64,
            unprotected=unprotected,
            ciphertext=parsed["ciphertext"],
            iv=parsed.get("iv"),
            tag=parsed["tag"],
            aad=parsed.get("aad"),
        )
        all_h = protected.keys() | unprotected.keys()
        for recip in recips:
            if recip.header and recip.header.keys() & all_h:
                raise ValidationError("Invalid JWE: duplicate header")
            inst.add_recipient(recip)

        return inst

    def serialize(self) -> dict:
        """Serialize the JWE envelope to a mapping."""
        if self.protected_b64 is None:
            raise ValidationError("Missing protected: use set_protected")
        if self.ciphertext is None:
            raise ValidationError("Missing ciphertext for JWE")
        if self.iv is None:
            raise ValidationError("Missing iv (nonce) for JWE")
        if self.tag is None:
            raise ValidationError("Missing tag for JWE")
        env = OrderedDict()
        env["protected"] = self.protected_b64
        if self.unprotected:
            env["unprotected"] = self.unprotected
        env["iv"] = b64url(self.iv)
        env["ciphertext"] = b64url(self.ciphertext)
        env["tag"] = b64url(self.tag)
        if self.aad:
            env["aad"] = b64url(self.aad)
        return env

    def to_json(self) -> str:
        """Serialize the JWE envelope to a JSON string."""
        return json.dumps(self.serialize())

    def add_recipient(self, recip: JweRecipient):
        """Add a recipient to the JWE envelope."""
        self._recipients.append(recip)

    def set_protected(self, protected: Mapping[str, Any], auto_flatten: bool = True):
        """Set the protected headers of the JWE envelope.

        This method must be called after adding the message recipients,
        or with the recipients pre-encoded and added to the headers.
        """
        protected = OrderedDict(protected.items())
        have_recips = IDENT_RECIPIENTS in protected or IDENT_ENC_KEY in protected
        if not have_recips:
            recipients = [recip.serialize() for recip in self._recipients]
            if auto_flatten and len(recipients) == 1:
                protected[IDENT_ENC_KEY] = recipients[0]["encrypted_key"]
                if "header" in recipients[0]:
                    protected[IDENT_HEADER] = recipients[0]["header"]
            elif recipients:
                protected[IDENT_RECIPIENTS] = recipients
            else:
                raise ValidationError("Missing message recipients")
        self.protected_b64 = b64url(json.dumps(protected))

    @property
    def protected_bytes(self) -> bytes:
        """Access the protected data encoded as bytes.

        This value is used in the additional authenticated data when encrypting.
        """
        return (
            self.protected_b64.encode("utf-8")
            if self.protected_b64 is not None
            else None
        )

    def set_payload(self, ciphertext: bytes, iv: bytes, tag: bytes, aad: bytes = None):
        """Set the payload of the JWE envelope."""
        self.ciphertext = ciphertext
        self.iv = iv
        self.tag = tag
        self.aad = aad

    def recipients(self) -> Iterable[JweRecipient]:
        """Accessor for an iterator over the JWE recipients.

        The headers for each recipient include protected and unprotected headers from the
        outer envelope.
        """
        header = self.protected.copy()
        header.update(self.unprotected)
        for recip in self._recipients:
            if recip.header:
                recip_h = header.copy()
                recip_h.update(recip.header)
                yield JweRecipient(encrypted_key=recip.encrypted_key, header=recip_h)
            else:
                yield JweRecipient(encrypted_key=recip.encrypted_key, header=header)



import datetime

from aries_cloudagent.messaging.models.base import BaseModel, BaseModelSchema
from aries_cloudagent.messaging.valid import UUIDFour
from marshmallow import fields, EXCLUDE
from marshmallow.exceptions import ValidationError
from typing import List

from ..utils.regex import MYDATA_DID


class DataAgreementResponseBody(BaseModel):
    """
    Data agreement response body model class
    """
    class Meta:
        schema_class = "DataAgreementResponseBodySchema"

    def __init__(self,
                 *,
                 data_agreement_id: str,
                 **kwargs):
        super().__init__(**kwargs)
        self.data_agreement_id = data_agreement_id

class DataAgreementResponseBodySchema(BaseModelSchema):
    """
    Data agreement response schema class
    """
    class Meta:
        model_class = DataAgreementResponseBody
        unknown = EXCLUDE

    data_agreement_id = fields.Str(
        data_key="data_agreement_id",
        example="45988dd2-62b9-4ede-8189-fb99c64b42d1",
        description="Data agreement identifier"
    )


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mydata_did',
 'mydata_did.patched_protocols',
 'mydata_did.patched_protocols.issue_credential',
 'mydata_did.patched_protocols.issue_credential.v1_0',
 'mydata_did.patched_protocols.issue_credential.v1_0.handlers',
 'mydata_did.patched_protocols.issue_credential.v1_0.messages',
 'mydata_did.patched_protocols.issue_credential.v1_0.messages.inner',
 'mydata_did.patched_protocols.issue_credential.v1_0.models',
 'mydata_did.patched_protocols.present_proof',
 'mydata_did.patched_protocols.present_proof.v1_0',
 'mydata_did.patched_protocols.present_proof.v1_0.handlers',
 'mydata_did.patched_protocols.present_proof.v1_0.messages',
 'mydata_did.patched_protocols.present_proof.v1_0.messages.inner',
 'mydata_did.patched_protocols.present_proof.v1_0.models',
 'mydata_did.patched_protocols.present_proof.v1_0.util',
 'mydata_did.v1_0',
 'mydata_did.v1_0.decorators',
 'mydata_did.v1_0.handlers',
 'mydata_did.v1_0.messages',
 'mydata_did.v1_0.models',
 'mydata_did.v1_0.models.exchange_records',
 'mydata_did.v1_0.routes',
 'mydata_did.v1_0.routes.maps',
 'mydata_did.v1_0.routes.openapi',
 'mydata_did.v1_0.utils',
 'mydata_did.v1_0.utils.did',
 'mydata_did.v1_0.utils.jsonld',
 'mydata_did.v1_0.utils.wallet']

package_data = \
{'': ['*']}

install_requires = \
['MarkupSafe==2.0.1',
 'PyJWT>=2.4.0,<3.0.0',
 'acapy-patched==0.5.6.dev1',
 'dexa-sdk==0.1.1',
 'py-multibase>=1.0.3,<2.0.0',
 'py-solc-x>=1.1.1,<2.0.0',
 'python3-indy>=1.16.0,<2.0.0',
 'semver>=2.13.0,<3.0.0',
 'validators>=0.20.0,<0.21.0']

setup_kwargs = {
    'name': 'acapy-mydata-did-protocol',
    'version': '0.1.4',
    'description': 'ACA-Py plugin for MyData DID DIDComm protcol',
    'long_description': '# ACA-Py plugin for MyData DID DIDComm protcol\n\n## Acknowledgements\n\nThis repository was originally created as as a deliverable for Automated Data Agreement (ADA) Project. ADA project is part of NGI-eSSIF-Lab project that has received funding from the European Union’s Horizon 2020 research and innovation programme under grant agreement No 871932.\n\nThe lead developer to this project is iGrant.io (Sweden), supported by Linaltec (Sweden) and PrivacyAnt (Finland).\n\n## ACA-Py Version Compatibility\n\nThis plugin is compatible with ACA-Py version 0.5.6.\n\n## Installation\n\nRequirements:\n- Python 3.6 or higher\n- ACA-Py 0.5.6\n\n### Setup Aries Cloud Agent - Python\n\nIf you already have an existing installation of ACA-Py, you can skip these steps\nand move on to [plugin installation](#plugin-installation). It is also worth\nnoting that this is not the only way to setup an ACA-Py instance. For more setup\nconfigurations, see the [Aries Cloud Agent - Python\nrepository](https://github.com/hyperledger/aries-cloudagent-python).\n\nFirst, prepare a virtual environment:\n```sh\n$ python3 -m venv env\n$ source env/bin/activate\n```\n\nInstall ACA-Py 0.5.6 into the virtual environment:\n```sh\n$ pip install aries-cloudagent==0.5.6\n```\n\n### Plugin Installation\n\nInstall this plugin into the virtual environment:\n\n```sh\n$ pip install acapy-mydata-did-protocol\n```\n\n**Note:** Depending on your version of `pip`, you may need to drop or add \n`#egg=mydata_did` to install the plugin with the above command.\n\n### Plugin Loading\nStart up ACA-Py with the plugin parameter:\n```sh\n$ aca-py start \\\n    -it http 0.0.0.0 8002 \\\n    -ot http \\\n    -e "http://localhost:8002/" \\\n    --label "Agent" \\\n    --admin 0.0.0.0 8001 \\\n    --admin-insecure-mode \\\n    --auto-accept-requests \\\n    --auto-ping-connection \\\n    --auto-respond-credential-offer \\\n    --auto-respond-credential-request \\\n    --auto-store-credential \\\n    --auto-respond-presentation-proposal \\\n    --auto-respond-presentation-request \\\n    --auto-verify-presentation \\\n    --genesis-url https://indy.igrant.io/genesis \\\n    --wallet-type indy \\\n    --wallet-name "agent_wallet" \\\n    --log-level info \\\n    --wallet-key "wallet@123" \\\n    --plugin "mydata_did"\n```\n\n## Licensing\n\nCopyright (c) 2021-23 LCubed AB (iGrant.io), Sweden\n\nLicensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the License.\n\nYou may obtain a copy of the License at https://www.apache.org/licenses/LICENSE-2.0.\n\nUnless required by applicable law or agreed to in writing, software distributed under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the LICENSE for the specific language governing permissions and limitations under the License.\n',
    'author': 'George J Padayatti',
    'author_email': 'george.padayatti@igrant.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/decentralised-dataexchange/acapy-mydata-did-protocol',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from os import environ
from typing import Any
import typing

from marshmallow import fields, validate

from aries_cloudagent.messaging.models.base_record import BaseExchangeRecord, BaseExchangeSchema
from aries_cloudagent.messaging.valid import UUIDFour


class DataAgreementCRUDDIDCommTransaction(BaseExchangeRecord):
    """
    Transaction record model class for Data Agreement CRUD didcomm message lifecycle.
    for e.g. create-data-agreement and create-data-agreement-response messages are part of
    create-data-agreement didcomm message lifecycle.

    Model class will create a transaction record adhering to the schema specified and
    store the record in the wallet.

    Each transaction record will have 2 tags: thread_id and message_type.

    1. thread_id - identifier of the initial message that triggered the transaction.
    2. message_type - identifier of the message family that triggered the transaction 
                        for .e.g. create-data-agreement
    """
    class Meta:
        # Schema class
        schema_class = "DataAgreementCRUDDIDCommTransactionSchema"

    RECORD_TYPE = "data_agreement_crud_didcomm_transaction"
    RECORD_ID_NAME = "da_crud_didcomm_tx_id"
    WEBHOOK_TOPIC = None
    TAG_NAMES = {"~thread_id", "~message_type", "~connection_id"}

    # Message family for the transaction record
    # Create Data Agreement message family
    MESSAGE_TYPE_CREATE_DATA_AGREEMENT = "create-data-agreement"
    # Read Data Agreement message family
    MESSAGE_TYPE_READ_DATA_AGREEMENT = "read-data-agreement"
    # Update Data Agreement message family
    MESSAGE_TYPE_UPDATE_DATA_AGREEMENT = "update-data-agreement"
    # Delete Data Agreement message family
    MESSAGE_TYPE_DELETE_DATA_AGREEMENT = "delete-data-agreement"

    def __init__(
        self,
        *,
        da_crud_didcomm_tx_id: str = None,
        thread_id: str = None,
        message_type: str = None,
        messages_list: typing.List[dict] = None,
        connection_id: str = None,
        **kwargs
    ):
        super().__init__(da_crud_didcomm_tx_id, **kwargs)
        self.thread_id = thread_id
        self.message_type = message_type
        self.messages_list = messages_list
        self.connection_id = connection_id

    @property
    def da_crud_didcomm_tx_id(self) -> str:
        """
        Returns transaction record identifier.
        """
        return self._id

    @property
    def record_value(self) -> dict:
        """Accessor for JSON record value generated for this transaction record."""
        return {
            prop: getattr(self, prop)
            for prop in (
                "thread_id",
                "message_type",
                "messages_list",
                "connection_id",
            )
        }

    def __eq__(self, other: Any) -> bool:
        """Comparison between records."""
        return super().__eq__(other)


class DataAgreementCRUDDIDCommTransactionSchema(BaseExchangeSchema):

    class Meta:
        # Model class
        model_class = DataAgreementCRUDDIDCommTransaction
    
    # Transaction record id
    da_crud_didcomm_tx_id = fields.Str(
        required=True,
        description="Transaction record id",
        example=UUIDFour.EXAMPLE,
    )
    
    # Parent message id of the transaction record
    thread_id = fields.Str(
        required=False,
        description="Parent message identifier",
        example=UUIDFour.EXAMPLE,
    )
    # Message type of the transaction record
    message_type = fields.Str(
        required=False,
        description="DIDComm message family",
        example="create-data-agreement",
    )
    # List of messages in the transaction record
    messages_list = fields.List(fields.Dict(), required=False, description="Messages list")

    # Connection identifier
    connection_id = fields.Str(
        required=False,
        description="Connection identifier",
        example=UUIDFour.EXAMPLE,
    )



from aries_cloudagent.messaging.base_handler import BaseHandler, BaseResponder, RequestContext
from aries_cloudagent.storage.record import StorageRecord
from aries_cloudagent.wallet.base import BaseWallet
from aries_cloudagent.wallet.indy import IndyWallet

from ..messages.problem_report import DataAgreementTerminationProblemReport
from ..models.exchange_records.data_agreement_record import DataAgreementV1Record
from ..manager import ADAManager

from ...patched_protocols.issue_credential.v1_0.models.credential_exchange import (
    V10CredentialExchange
)

from ...patched_protocols.present_proof.v1_0.models.presentation_exchange import (
    V10PresentationExchange
)

import json


class DataAgreementTerminationProblemReportHandler(BaseHandler):
    """
    Handler for data agreement termination problem report.
    """

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """
        Message handler logic for data agreement termination problem report.
        """

        # Assert if received message is of type DataAgreementTerminationProblemReport
        assert isinstance(context.message, DataAgreementTerminationProblemReport)

        self._logger.info(
            "Received data-agreement-termination/1.0/problem-report message: \n%s",
            json.dumps(context.message.serialize(), indent=4)
        )

        # Check if the connection is ready.
        if not context.connection_ready:
            self._logger.info(
                "Connection not active, skipping data-agreement-termination/1.0/problem-report handler: %s",
                context.message_receipt.sender_did,
            )
            return
        
        data_agreement_termination_problem_report = context.message
        

        # Query data agreement instance metadata

        # Wallet instance from request context
        wallet: IndyWallet = await context.inject(BaseWallet)

        # Initialize ADA manager
        ada_manager = ADAManager(context)

        # Fetch the data agreement instance metadata
        data_agreement_instance_metadata_records = await ada_manager.query_data_agreement_instance_metadata(
            tag_query={
                'data_agreement_id': data_agreement_termination_problem_report.data_agreement_id,
            }
        )

        # Check if there is a data agreement instance metadata record
        if not data_agreement_instance_metadata_records:
            self._logger.info(
                "Data agreement not found; Failed to handle termination problem report for data agreement: %s",
                data_agreement_termination_problem_report.data_agreement_id,
            )
            return
        
        if len(data_agreement_instance_metadata_records) > 1:
            self._logger.info(
                "Duplicate data agreement records found; Failed to handle termination problem report for data agreement: %s",
                data_agreement_termination_problem_report.data_agreement_id,
            )
            return
        
        data_agreement_instance_metadata_record: StorageRecord = data_agreement_instance_metadata_records[0]

        if data_agreement_instance_metadata_record.tags.get("method_of_use") == DataAgreementV1Record.METHOD_OF_USE_DATA_SOURCE:
            # Check method of use and fetch appropriate exchange record

            # Fetch exchange record (credential exchange if method of use is "data-source")
            tag_filter = {}
            post_filter = {
                "data_agreement_id": data_agreement_termination_problem_report.data_agreement_id
            }
            records = await V10CredentialExchange.query(context, tag_filter, post_filter)

            if not records:
                self._logger.info(
                    "Credential exchange record not found; Failed to handle termination problem report for data agreement: %s",
                    data_agreement_termination_problem_report.data_agreement_id,
                )
                return
            
            if len(records) > 1:
                self._logger.info(
                    "Duplicate credential exchange records found; Failed to handle termination problem report for data agreement: %s",
                    data_agreement_termination_problem_report.data_agreement_id,
                )
                return
            
            cred_ex_record: V10CredentialExchange = records[0]

            # Update exchange record data agreement status
            cred_ex_record.data_agreement_status = V10CredentialExchange.DATA_AGREEMENT_PROBLEM_REPORT

            # Add problem report to credential exchange record
            cred_ex_record.data_agreement_problem_report = data_agreement_termination_problem_report.serialize()

            # Save the credential exchange record
            await cred_ex_record.save(context)

        if data_agreement_instance_metadata_record.tags.get("method_of_use") == DataAgreementV1Record.METHOD_OF_USE_DATA_USING_SERVICE:
            # Implement data agreement termination problem report handler for data using service

            # Check method of use and fetch appropriate exchange record

            # Fetch exchange record (presentation exchange if method of use is "data-using-service")
            tag_filter = {}
            post_filter = {
                "data_agreement_id": data_agreement_termination_problem_report.data_agreement_id
            }
            records = await V10PresentationExchange.query(context, tag_filter, post_filter)

            if not records:
                self._logger.info(
                    "Presentation exchange record not found; Failed to handle termination problem report for data agreement: %s",
                    data_agreement_termination_problem_report.data_agreement_id,
                )
                return
            
            if len(records) > 1:
                self._logger.info(
                    "Duplicate presentation exchange records found; Failed to handle termination problem report for data agreement: %s",
                    data_agreement_termination_problem_report.data_agreement_id,
                )
                return
            
            pres_ex_record: V10PresentationExchange = records[0]

            # Update exchange record data agreement status
            pres_ex_record.data_agreement_status = V10PresentationExchange.DATA_AGREEMENT_PROBLEM_REPORT

            # Add problem report to presentation exchange record
            pres_ex_record.data_agreement_problem_report = data_agreement_termination_problem_report.serialize()

            # Save the presentation exchange record
            await pres_ex_record.save(context)


        






from aries_cloudagent.messaging.agent_message import AgentMessage, AgentMessageSchema
from marshmallow import EXCLUDE, fields
from ..message_types import DATA_AGREEMENT_NEGOTIATION_RECEIPT, PROTOCOL_PACKAGE
from ..models.data_agreement_negotiation_receipt_model import (
    DataAgreementNegotiationReceiptBody,
    DataAgreementNegotiationReceiptBodySchema
)

HANDLER_CLASS = (
    f"{PROTOCOL_PACKAGE}.handlers"
    ".da_negotiation_receipt_handler.DataAgreementNegotiationReceiptMessageHandler"
)


class DataAgreementNegotiationReceiptMessage(AgentMessage):
    """
    Message class for data agreement negotiation receipt message.
    """

    class Meta:

        # Message type
        message_type = DATA_AGREEMENT_NEGOTIATION_RECEIPT

        # Schema class
        schema_class = "DataAgreementNegotiationReceiptMessageSchema"

        # Handler class
        handler_class = HANDLER_CLASS

    def __init__(
        self,
        *,
        body: DataAgreementNegotiationReceiptBody,
        **kwargs
    ):
        """
        Initialize data agreement negotiation receipt message.
        """
        super().__init__(**kwargs)

        # Message body
        self.body = body


class DataAgreementNegotiationReceiptMessageSchema(AgentMessageSchema):
    """
    Schema class for data agreement negotiation accept message
    """

    class Meta:
        # Model class
        model_class = DataAgreementNegotiationReceiptMessage

        # Unknown fields to exclude from the schema (handled by marshmallow)
        unknown = EXCLUDE

    # Message body
    body = fields.Nested(
        DataAgreementNegotiationReceiptBodySchema,
        required=True
    )



from aries_cloudagent.messaging.base_handler import BaseHandler, BaseResponder, RequestContext
from dexa_sdk.managers.ada_manager import V2ADAManager
from ..messages.da_negotiation_receipt import DataAgreementNegotiationReceiptMessage


class DataAgreementNegotiationReceiptMessageHandler(BaseHandler):
    """Data agreement negotiation receipt message handler."""

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """Handle function

        Args:
            context (RequestContext): Request context.
            responder (BaseResponder): Responder.
        """

        # Assert the message type.
        assert isinstance(context.message, DataAgreementNegotiationReceiptMessage)

        # Initialize manager.
        mgr = V2ADAManager(context)

        # Process read-did message
        await mgr.process_read_did_message(context.message, context.message_receipt)



from marshmallow import EXCLUDE, fields

from aries_cloudagent.messaging.agent_message import AgentMessage, AgentMessageSchema
from aries_cloudagent.messaging.models.base import BaseModel, BaseModelSchema

from ..message_types import PROTOCOL_PACKAGE, READ_DID
from ..utils.regex import MYDATA_DID

HANDLER_CLASS = (
    f"{PROTOCOL_PACKAGE}.handlers"
    ".read_did_handler.ReadDIDHandler"
)

class ReadDIDMessageBody(BaseModel):
    """
    Read DID message body class
    """    
    class Meta:

        # Schema class
        schema_class = "ReadDIDMessageBodySchema"
    
    def __init__(self, *, did: str, **kwargs):
        """
        Initialize ReadDIDMessageBody instance

        Args:
            did: The DID to be read
        """

        super().__init__(**kwargs)

        # The DID to be read
        self.did = did

class ReadDIDMessageBodySchema(BaseModelSchema):
    """
    Read DID message body schema class
    """
    class Meta:
        # Message body model
        model_class = ReadDIDMessageBody

        # Unknown fields to exclude from the schema
        unknown = EXCLUDE
    
    # The DID to be read
    did = fields.Str(data_key="did", **MYDATA_DID)

class ReadDIDMessage(AgentMessage):
    """
    Message class for reading a DID.
    """
    class Meta:

        # Handler class that can handle this message
        handler_class = HANDLER_CLASS

        # Message type
        message_type = READ_DID

        # Message schema class
        schema_class = "ReadDIDMessageSchema"

    def __init__(self, *, from_did, to_did, created_time, body: ReadDIDMessageBody, **kwargs):
        """
        Initialize a ReadDIDMessage message instance.

        Args:
            from_did: Sender DID
            to_did: Reciepient DID
            created_time: Time the message was created
            body: Message body
        """

        super().__init__(**kwargs)

        # Set attributes
        self.from_did = from_did
        self.to_did = to_did
        self.created_time = created_time
        self.body = body

class ReadDIDMessageSchema(AgentMessageSchema):
    """
    Schema class for reading a DID.
    """
    
    class Meta:

        # The message class that this schema is for
        model_class = ReadDIDMessage

        # Unknown fields to exclude from the schema (handled by marshmallow)
        unknown = EXCLUDE
    
    # From DID
    from_did = fields.Str(data_key="from", **MYDATA_DID)

    # To DID
    to_did = fields.Str(data_key="to", **MYDATA_DID)

    # Created time
    created_time = fields.Str(data_key="created_time")

    # Message body
    body = fields.Nested(ReadDIDMessageBodySchema, required=True)



import base64
import logging
import json
import time
import typing

import aiohttp

from aries_cloudagent.config.injection_context import InjectionContext
from aries_cloudagent.connections.models.connection_record import ConnectionRecord
from aries_cloudagent.connections.models.connection_target import ConnectionTarget
from aries_cloudagent.messaging.responder import BaseResponder
from aries_cloudagent.core.dispatcher import DispatcherResponder
from aries_cloudagent.transport.inbound.receipt import MessageReceipt
from aries_cloudagent.core.error import BaseError
from aries_cloudagent.storage.base import BaseStorage, StorageRecord
from aries_cloudagent.storage.indy import IndyStorage
from aries_cloudagent.storage.error import (
    StorageNotFoundError,
    StorageDuplicateError,
    StorageError
)
from aries_cloudagent.wallet.indy import IndyWallet
from aries_cloudagent.wallet.base import BaseWallet
from aries_cloudagent.transport.pack_format import PackWireFormat
from aries_cloudagent.transport.wire_format import BaseWireFormat
from aries_cloudagent.messaging.decorators.transport_decorator import TransportDecorator
from aries_cloudagent.protocols.connections.v1_0.manager import (
    ConnectionManager,
)
from .messages.read_did import ReadDIDMessage, ReadDIDMessageBody
from .messages.read_did_response import ReadDIDResponseMessage, ReadDIDResponseMessageSchema
from .messages.problem_report import (
    MyDataDIDProblemReportMessage,
    MyDataDIDProblemReportMessageReason
)
from .messages.json_ld_processed import JSONLDProcessedMessage
from .messages.json_ld_processed_response import JSONLDProcessedResponseMessage
from .messages.json_ld_problem_report import JSONLDProblemReport, JSONLDProblemReportReason
from .models.diddoc_model import (
    MyDataDIDResponseBody,
    MyDataDIDDoc,
)
from .models.json_ld_processed_response_model import JSONLDProcessedResponseBody
from .models.json_ld_processed_model import JSONLDProcessedBody
from .utils.did.mydata_did import DIDMyData
from .utils.wallet.key_type import KeyType
from .utils.jsonld.create_verify_data import create_verify_data


class ADAManagerError(BaseError):
    """ADA manager error"""


class ADAManager:

    # Record for storing data agreement instance metadata (client)
    RECORD_TYPE_DATA_AGREEMENT_INSTANCE_METADATA = "data_agreement_instance_metadata"

    # Record for keeping track of DIDs that are registered in the DID registry (MyData DID registry)
    RECORD_TYPE_MYDATA_DID_REGISTRY_DID_INFO = "mydata_did_registry_did_info"

    # Record for keeping metadata about data agreement QR codes (client)
    RECORD_TYPE_DATA_AGREEMENT_QR_CODE_METADATA = "data_agreement_qr_code_metadata"

    # Temporary record for keeping personal data of unpublished (or draft) data agreements
    RECORD_TYPE_TEMPORARY_DATA_AGREEMENT_PERSONAL_DATA = "temporary_data_agreement_personal_data"

    # Record for data controller details
    RECORD_TYPE_DATA_CONTROLLER_DETAILS = "data_controller_details"

    # Record for existing connection details.
    RECORD_TYPE_EXISTING_CONNECTION = "existing_connection"

    DATA_AGREEMENT_RECORD_TYPE = "dataagreement_record"

    def __init__(self, context: InjectionContext) -> None:
        self._context = context
        self._logger = logging.getLogger(__name__)

    @property
    def context(self) -> InjectionContext:
        return self._context

    async def process_read_did_message(self,
                                       read_did_message: ReadDIDMessage,
                                       receipt: MessageReceipt):
        """
        Process read-did DIDComm message
        """

        # Storage instance from context
        storage: IndyStorage = await self.context.inject(BaseStorage)

        # Responder instance from context
        responder: DispatcherResponder = await self.context.inject(BaseResponder, required=False)

        # From and To DIDs of the recieved message
        create_did_message_from_did: DIDMyData = DIDMyData.from_public_key_b58(
            receipt.sender_verkey, key_type=KeyType.ED25519)
        create_did_message_to_did: DIDMyData = DIDMyData.from_public_key_b58(
            receipt.recipient_verkey, key_type=KeyType.ED25519)

        # From and To DIDs for the response messages
        response_message_from_did = create_did_message_to_did
        response_message_to_did = create_did_message_from_did

        mydata_did_registry_did_info_record = None
        try:

            # Fetch DID from wallet
            mydata_did_registry_did_info_record = await storage.search_records(
                type_filter=ADAManager.RECORD_TYPE_MYDATA_DID_REGISTRY_DID_INFO,
                tag_query={"did": read_did_message.body.did}
            ).fetch_single()

        except (StorageNotFoundError, StorageDuplicateError):
            # Send problem-report message.

            mydata_did_problem_report = MyDataDIDProblemReportMessage(
                problem_code=MyDataDIDProblemReportMessageReason.DID_NOT_FOUND.value,
                explain="DID not found.",
                from_did=response_message_from_did.did,
                to_did=response_message_to_did.did,
                created_time=round(time.time() * 1000)
            )

            # Assign thread id
            mydata_did_problem_report.assign_thread_id(
                thid=read_did_message._id)

            if responder:
                await responder.send_reply(mydata_did_problem_report)

            return

        # Send read-did-response message
        read_did_response_message = ReadDIDResponseMessage(
            from_did=response_message_from_did.did,
            to_did=response_message_to_did.did,
            created_time=round(time.time() * 1000),
            body=MyDataDIDResponseBody(
                did_doc=MyDataDIDDoc.from_json(
                    mydata_did_registry_did_info_record.value),
                version=mydata_did_registry_did_info_record.tags.get(
                    "version"),
                status=mydata_did_registry_did_info_record.tags.get("status")
            )
        )

        # Assign thread id
        read_did_response_message.assign_thread_id(
            thid=read_did_message._id)

        if responder:
            await responder.send_reply(read_did_response_message)

    async def process_read_did_response_message(
        self,
        read_did_response_message: ReadDIDResponseMessage,
        receipt: MessageReceipt
    ):
        """
        Process read-did-response DIDComm message
        """

        pass

    async def send_read_did_message(self, did: str):
        """
        Send read-did DIDComm message
        """

        pass

    async def store_data_agreement_instance_metadata(
        self,
        *,
        data_agreement_id: str = None,
        data_agreement_template_id: str = None,
        method_of_use: str = None,
        data_exchange_record_id: str = None
    ) -> None:
        """Store data agreement instance metadata"""

        # Fetch storage from context
        storage: IndyStorage = await self.context.inject(BaseStorage)

        data_instance_metadata_record = StorageRecord(
            self.RECORD_TYPE_DATA_AGREEMENT_INSTANCE_METADATA,
            data_agreement_id,
            {
                "data_agreement_id": data_agreement_id,
                "data_agreement_template_id": data_agreement_template_id,
                "method_of_use": method_of_use,
                "data_exchange_record_id": data_exchange_record_id
            }
        )

        await storage.add_record(data_instance_metadata_record)

    async def delete_data_agreement_instance_metadata(self, *, tag_query: dict = None) -> None:
        """Delete data agreement instance metadata"""

        # Fetch storage from context
        storage: IndyStorage = await self.context.inject(BaseStorage)

        storage_records = await storage.search_records(
            type_filter=self.RECORD_TYPE_DATA_AGREEMENT_INSTANCE_METADATA,
            tag_query=tag_query
        ).fetch_all()

        for storage_record in storage_records:
            await storage.delete_record(storage_record)

    async def query_data_agreement_instance_metadata(self, *, tag_query: dict = None) -> typing.List[StorageRecord]:
        """Query data agreement instance metadata"""

        # Fetch storage from context
        storage: IndyStorage = await self.context.inject(BaseStorage)

        storage_records = await storage.search_records(
            type_filter=self.RECORD_TYPE_DATA_AGREEMENT_INSTANCE_METADATA,
            tag_query=tag_query
        ).fetch_all()

        return storage_records

    async def resolve_remote_mydata_did(self, *, mydata_did: str) -> MyDataDIDResponseBody:
        """Resolve remote MyData DID"""

        # Initialize DID MyData
        mydata_did = DIDMyData.from_did(mydata_did)

        # Fetch wallet from context
        wallet: IndyWallet = await self.context.inject(BaseWallet)

        # Get pack format from context
        pack_format: PackWireFormat = await self.context.inject(BaseWireFormat)

        # Fetch connection record marked as MyData DID registry
        connection_record, err = await self.fetch_mydata_did_registry_connection_record()
        if err:
            raise ADAManagerError(
                "Failed to fetch MyData DID registry connection record")

        # Construct read-did message
        # from_did
        pairwise_local_did_record = await wallet.get_local_did(connection_record.my_did)
        from_did = DIDMyData.from_public_key_b58(
            pairwise_local_did_record.verkey, key_type=KeyType.ED25519)

        # to_did
        to_did = DIDMyData.from_public_key_b58(
            connection_record.their_did, key_type=KeyType.ED25519)

        # Create read-did message
        read_did_message = ReadDIDMessage(
            from_did=from_did.did,
            to_did=to_did.did,
            created_time=round(time.time() * 1000),
            body=ReadDIDMessageBody(
                did=mydata_did.did
            )
        )

        # Add transport decorator
        read_did_message._decorators["transport"] = TransportDecorator(
            return_route="all"
        )

        # Initialise connection manager
        connection_manager = ConnectionManager(self.context)

        # Fetch connection targets
        connection_targets = await connection_manager.fetch_connection_targets(connection_record)

        if len(connection_targets) == 0:
            raise ADAManagerError("No connection targets found")

        connection_target: ConnectionTarget = connection_targets[0]

        # Pack message
        packed_message = await pack_format.pack(
            context=self.context,
            message_json=read_did_message.serialize(as_string=True),
            recipient_keys=connection_target.recipient_keys,
            routing_keys=None,
            sender_key=connection_target.sender_key,
        )

        headers = {
            "Content-Type": "application/ssi-agent-wire"
        }
        async with aiohttp.ClientSession(headers=headers) as session:
            async with session.post(connection_target.endpoint, data=packed_message) as response:
                if response.status != 200:
                    raise ADAManagerError(
                        f"HTTP request failed with status code {response.status}")

                message_body = await response.read()

                unpacked = await wallet.unpack_message(message_body)

                (
                    message_json,
                    sender_verkey,
                    recipient_verkey,
                ) = unpacked

                message_json = json.loads(message_json)

                if "problem-report" in message_json["@type"]:
                    raise ADAManagerError(
                        f"Problem report received with problem-code:{message_json['problem-code']} and reason: {message_json['explain']}")

                if "read-did-response" in message_json["@type"]:
                    read_did_response_message: ReadDIDResponseMessage = \
                        ReadDIDResponseMessageSchema().load(message_json)

                    if read_did_response_message.body.status == "revoked":
                        raise ADAManagerError(
                            f"MyData DID {mydata_did.did} is revoked"
                        )

                    return read_did_response_message.body

    async def process_json_ld_processed_message(
        self,
        json_ld_processed_message: JSONLDProcessedMessage,
        receipt: MessageReceipt
    ) -> None:

        # Responder instance
        responder: DispatcherResponder = await self.context.inject(BaseResponder, required=False)

        # From and To MyData DIDs
        to_did: DIDMyData = DIDMyData.from_public_key_b58(
            receipt.sender_verkey, key_type=KeyType.ED25519)
        from_did: DIDMyData = DIDMyData.from_public_key_b58(
            receipt.recipient_verkey, key_type=KeyType.ED25519)

        try:
            # Base64 decode data_base64
            data_base64_decoded = base64.b64decode(
                json_ld_processed_message.body.data_base64)

            # JSON load data_base64
            data_json = json.loads(data_base64_decoded)

            # Base64 decode signature_options_base64
            signature_options_base64_decoded = base64.b64decode(
                json_ld_processed_message.body.signature_options_base64)

            # JSON load signature_options_base64
            signature_options_json = json.loads(
                signature_options_base64_decoded)

            # verify_data function
            framed, combine_hash = create_verify_data(
                data_json, signature_options_json, json_ld_processed_message.body.proof_chain)

            # Base64 encode framed
            framed_base64_encoded = base64.b64encode(
                json.dumps(framed).encode("utf-8")).decode("utf-8")

            # Base64 encode combine_hash
            combine_hash_base64_encoded = base64.b64encode(
                combine_hash.encode("utf-8")).decode("utf-8")

            # Construct JSONLD Processed Response Message
            json_ld_processed_response_message = JSONLDProcessedResponseMessage(
                from_did=from_did.did,
                to_did=to_did.did,
                created_time=round(time.time() * 1000),
                body=JSONLDProcessedResponseBody(
                    framed_base64=framed_base64_encoded,
                    combined_hash_base64=combine_hash_base64_encoded
                )
            )

            if responder:
                await responder.send_reply(
                    json_ld_processed_response_message,
                    connection_id=self.context.connection_record.connection_id
                )

        except Exception as err:
            # Send problem report
            json_ld_problem_report_message = JSONLDProblemReport(
                problem_code=JSONLDProblemReportReason.INVALID_INPUT.value,
                explain=str(err),
                from_did=from_did.did,
                to_did=to_did.did,
                created_time=round(time.time() * 1000)
            )

            if responder:
                await responder.send_reply(
                    json_ld_problem_report_message,
                    connection_id=self.context.connection_record.connection_id
                )

    async def send_json_ld_processed_message(
        self,
        *,
        connection_id: str,
        data: dict,
        signature_options: dict,
        proof_chain: bool
    ) -> None:
        """Send JSON-LD Processed Message to remote agent."""

        # Responder instance
        responder: DispatcherResponder = await self.context.inject(BaseResponder, required=False)

        try:

            # Retrieve connection record by id
            connection_record: ConnectionRecord = await ConnectionRecord.retrieve_by_id(
                self.context,
                connection_id
            )

        except StorageError as err:

            raise ADAManagerError(
                f"Failed to retrieve connection record: {err}"
            )

        # From and to mydata dids
        from_did: DIDMyData = DIDMyData.from_public_key_b58(
            connection_record.my_did, key_type=KeyType.ED25519
        )
        to_did: DIDMyData = DIDMyData.from_public_key_b58(
            connection_record.their_did, key_type=KeyType.ED25519
        )

        # Base64 encode data
        data_base64 = base64.b64encode(json.dumps(
            data).encode("utf-8")).decode("utf-8")

        # Base64 encode signature_options
        signature_options_base64 = base64.b64encode(json.dumps(
            signature_options).encode("utf-8")).decode("utf-8")

        # Construct JSONLD Processed Message
        json_ld_processed_message = JSONLDProcessedMessage(
            from_did=from_did.did,
            to_did=to_did.did,
            created_time=round(time.time() * 1000),
            body=JSONLDProcessedBody(
                data_base64=data_base64,
                signature_options_base64=signature_options_base64,
                proof_chain=proof_chain
            )
        )

        # Send JSONLD Processed Message
        if responder:
            await responder.send_reply(
                json_ld_processed_message,
                connection_id=connection_record.connection_id
            )



from marshmallow import fields
from aries_cloudagent.messaging.models.base import BaseModel, BaseModelSchema
from aries_cloudagent.messaging.valid import UUIDFour


class JSONLDProcessedBody(BaseModel):
    """json-ld/1.0/processed-data message body"""

    class Meta:
        """json-ld/1.0/processed-data message body metadata"""

        schema_class = "JSONLDProcessedBodySchema"

    def __init__(
        self,
        *,
        data_base64: str = None,
        signature_options_base64: str = None,
        proof_chain: bool = False,
        **kwargs
    ):
        """json-ld/1.0/processed-data message body init"""

        super().__init__(**kwargs)

        # Set attributes
        self.data_base64 = data_base64
        self.signature_options_base64 = signature_options_base64
        self.proof_chain = proof_chain


class JSONLDProcessedBodySchema(BaseModelSchema):
    """json-ld/1.0/processed-data message body schema"""

    class Meta:
        """json-ld/1.0/processed-data message body schema metadata"""

        model_class = JSONLDProcessedBody

    # JSON-LD input data
    data_base64 = fields.Str(
        description="JSON-LD input data",
    )

    # Signature options
    signature_options_base64 = fields.Str(
        description="Signature options",
    )

    # Proof chain
    proof_chain = fields.Boolean(
        description="Proof chain",
    )



"""
DIDDoc utility methods.

Copyright 2017-2019 Government of Canada
Public Services and Procurement Canada - buyandsell.gc.ca

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import datetime
import semver

from urllib.parse import urlparse
from multibase import decode


if __name__ == "__main__":
    from mydata_did.v1_0.utils.regex import MYDATA_DID_PATTERN
else:
    from .regex import MYDATA_DID_PATTERN


def resource(ref: str, delimiter: str = None) -> str:
    """
    Extract the resource for an identifier.

    Given a (URI) reference, return up to its delimiter (exclusively), or all of it if
    there is none.

    Args:
        ref: reference
        delimiter: delimiter character
            (default None maps to '#', or ';' introduces identifiers)
    """

    return ref.split(delimiter if delimiter else "#")[0]


def derive_did_type(uri: str) -> str:
    mydata_did_pattern_match = MYDATA_DID_PATTERN.match(uri)
    if mydata_did_pattern_match:
        return mydata_did_pattern_match.group('did_type')
    return None


def canon_did(uri: str) -> str:
    """
    Convert a URI into a DID if need be, left-stripping 'did:mydata:' if present.

    Args:
        uri: input URI or DID

    Raises:
        ValueError: for invalid input.

    """

    if ok_did(uri):
        return uri

    if uri.startswith("did:mydata:"):
        mydata_did_pattern_match = MYDATA_DID_PATTERN.match(uri)
        if mydata_did_pattern_match:
            prefix_end = 13 if mydata_did_pattern_match.group(
                'did_type') else 11
            rv = uri[prefix_end:]
            if ok_did(rv):
                return rv
    raise ValueError(
        "Bad specification {} does not correspond to a MyData DID".format(uri)
    )


def canon_ref(did: str, ref: str, delimiter: str = None, did_type: str = None):
    """
    Given a reference in a DID document, return it in its canonical form of a URI.

    Args:
        did: DID acting as the identifier of the DID document
        ref: reference to canonicalize, either a DID or a fragment pointing to a
            location in the DID doc
        delimiter: delimiter character marking fragment (default '#') or
            introducing identifier (';') against DID resource
    """

    if not ok_did(did):
        raise ValueError(
            "Bad DID {} cannot act as DID document identifier".format(did))

    if ok_did(ref):  # e.g., LjgpST2rjsoxYegQDRm7EL
        return "did:mydata:{}".format(did) if not did_type else "did:mydata:{}:{}".format(did_type, did)

    if ok_did(resource(ref, delimiter)):  # e.g., LjgpST2rjsoxYegQDRm7EL#keys-1
        return "did:mydata:{}".format(ref) if not did_type else "did:mydata:{}:{}".format(did_type, did)

    if ref.startswith(
        "did:mydata:"
    ):  # e.g., did:mydata:LjgpST2rjsoxYegQDRm7EL, did:mydata:LjgpST2rjsoxYegQDRm7EL#3
        mydata_did_pattern_match = MYDATA_DID_PATTERN.match(
            resource(ref, delimiter))
        if mydata_did_pattern_match:
            prefix_end = 13 if mydata_did_pattern_match.group(
                'did_type') else 11
            rv = ref[prefix_end:]
            if ok_did(resource(rv, delimiter)):
                return ref
        raise ValueError(
            "Bad URI {} does not correspond to a MyData DID".format(ref))

    if urlparse(ref).scheme:  # e.g., https://example.com/messages/8377464
        return ref

    return "did:mydata:{}{}{}".format(did, delimiter if delimiter else "#", ref) if not did_type else "did:mydata:{}:{}{}{}".format(did_type, did, delimiter if delimiter else "#", ref)


def ok_did(token: str) -> bool:
    """
    Whether input token looks like a valid decentralized identifier.

    Args:
        token: candidate string

    Returns: whether input token looks like a valid schema identifier

    """
    try:
        return len(decode(token)) == 34 if token else False
    except ValueError:
        return False


def current_datetime_in_iso8601() -> str:
    """
    Return current datetime in ISO8601 format.
    """
    return str(datetime.datetime.utcnow().replace(tzinfo=datetime.timezone.utc).isoformat())


def str_to_bool(s: str) -> bool:
    """
    Convert a string to a boolean.

    Args:
        s: string to convert

    Returns: boolean value

    """

    if not isinstance(s, str):
        return False

    if s.lower() in ['true', 't', '1']:
        return True
    elif s.lower() in ['false', 'f', '0']:
        return False
    else:
        raise ValueError('Cannot convert string to boolean: {}'.format(s))


def bool_to_str(b: bool) -> str:
    """
    Convert a boolean to a string.

    Args:
        b: boolean to convert

    Returns: string value

    """
    return 'true' if b else 'false'


def int_to_semver_str(int_version: int) -> str:
    """
    Convert integer version to semver string.
    """
    return str(semver.VersionInfo(str(int_version)))


def comma_separated_str_to_list(s: str) -> list:
    """
    Convert a comma separated string to a list.

    Args:
        s: string to convert

    Returns: list value

    """
    return s.split(',')


def get_slices(page, page_size=10):
    """
    Get the start and end indices for the given page and page size.

    Args:
        page: page number
        page_size: page size

    Returns: start and end indices

    """
    start = (page - 1) * page_size

    end = start + page_size

    return start, end


if __name__ == "__main__":
    print(canon_did("did:mydata:0:z6MkfiSdYhnLnS6jfwSf2yS2CiwwjZGmFUFL5QbyL2Xu8z2E"))
